from crystal.model import Project, ResourceGroup
from crystal.tests.util.controls import click_button, TreeItem
from crystal.tests.util.server import extracted_project
from crystal.tests.util.wait import (
    first_child_of_tree_item_is_not_loading_condition, wait_for,
)
from crystal.tests.util.windows import NewRootUrlDialog, OpenOrCreateDialog
from unittest import skip


async def test_can_edit_name_of_root_url() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        home_url = 'https://xkcd.com/'
        home_g_pattern = 'https://xkcd.*/'
        assert home_g_pattern != home_url
        
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            ResourceGroup(project, 'Home Group', home_g_pattern)
            
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            home_ti = root_ti.GetFirstChild()
            assert home_ti is not None
            assert home_ti.Text.endswith('Home')
            
            home_ti.Expand()
            await wait_for(first_child_of_tree_item_is_not_loading_condition(home_ti))
            home_ti__home_ti = home_ti.find_child(home_url)
            assert home_ti__home_ti.Text.endswith('Home')
            
            home_g_ti = root_ti.find_child(home_g_pattern)
            assert home_g_ti.Text.endswith('Home Group')
            
            home_g_ti.Expand()
            await wait_for(first_child_of_tree_item_is_not_loading_condition(home_g_ti))
            home_g_ti__home_ti = home_g_ti.find_child(home_url)
            assert home_g_ti__home_ti.Text.endswith('Home')
            
            # Ensure can rename top-level RootResourceNode
            if True:
                home_ti.SelectItem()
                
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                nud = await NewRootUrlDialog.wait_for()
                
                # 1. Ensure prepopulates correct information
                # 2. Test: Cannot edit URL
                assert home_url == nud.url_field.Value
                assert 'Home' == nud.name_field.Value
                #assert None == ngd.source
                assert not nud.url_field.Enabled
                assert nud.name_field.HasFocus  # default focused field
                
                nud.name_field.Value = 'Home2'
                await nud.ok()
                home_ti = root_ti.GetFirstChild()
                assert home_ti is not None
                
                # Ensure selection did not change
                assert home_ti.IsSelected()
                assert home_ti.IsExpanded()
                
                # Ensure all copies of home node have updated name
                assert home_ti.Text.endswith('Home2')
                assert home_ti__home_ti.Text.endswith('Home2')
                assert home_g_ti__home_ti.Text.endswith('Home2')
            
            # Ensure can rename nested RootResourceNode
            if True:
                home_ti__home_ti.SelectItem()
                
                assert mw.edit_button.Enabled
                click_button(mw.edit_button)
                nud = await NewRootUrlDialog.wait_for()
                
                # Ensure prepopulates correct information
                assert home_url == nud.url_field.Value
                assert 'Home2' == nud.name_field.Value
                
                nud.name_field.Value = 'Home'
                await nud.ok()
                home_ti = root_ti.GetFirstChild()
                assert home_ti is not None
                
                # Ensure selection did not change
                assert home_ti.IsExpanded()
                assert home_ti__home_ti.IsSelected()
                
                # Ensure all copies of home node have updated name
                assert home_ti.Text.endswith('Home')
                assert home_ti__home_ti.Text.endswith('Home')
                assert home_g_ti__home_ti.Text.endswith('Home')


@skip('covered by: test_can_edit_name_of_root_url')
async def test_cannot_edit_url_of_root_url() -> None:
    pass



from crystal.util.xos import is_mac_os, is_windows
import ctypes
import os
import subprocess
import threading
from typing import Optional

_ES_CONTINUOUS = 0x80000000
_ES_SYSTEM_REQUIRED = 0x00000001  # prevents idle sleep
_ES_DISPLAY_REQUIRED = 0x00000002  # prevents screen sleep


class Caffeination:
    """
    Controls whether the system is prevented from idle sleeping.
    
    The screen is not kept awake during caffeination.
    """
    _lock = threading.Lock()
    _caffeine_count = 0
    _caffeinated = False
    _caffeinator = None  # type: Optional[subprocess.Popen]
    
    @classmethod
    def add_caffeine(cls) -> None:
        with cls._lock:
            old_caffeine_count = cls._caffeine_count  # capture
            cls._caffeine_count += 1
            if old_caffeine_count == 0:
                cls._set_caffeinated(True)
    
    @classmethod
    def remove_caffeine(cls) -> None:
        with cls._lock:
            cls._caffeine_count -= 1
            assert cls._caffeine_count >= 0
            new_caffeine_count = cls._caffeine_count  # capture
            if new_caffeine_count == 0:
                cls._set_caffeinated(False)
    
    @classmethod
    def _set_caffeinated(cls, caffeinated: bool) -> None:
        if caffeinated == cls._caffeinated:
            return
        if is_mac_os():
            if caffeinated:
                assert cls._caffeinator is None
                cls._caffeinator = subprocess.Popen(
                    [
                        'caffeinate',
                        # Wait for Crystal process to terminate
                        '-w', str(os.getpid()),
                        # No idle sleep
                        '-i',
                        # No sleep while on A/C power
                        '-s'
                    ],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL)
                assert cls._caffeinator.poll() is None, \
                    'caffeinate process terminated immediately unexpectedly'
            else:
                assert cls._caffeinator is not None
                cls._caffeinator.terminate()
                cls._caffeinator.wait()
                cls._caffeinator = None
        elif is_windows():
            # See: https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-setthreadexecutionstate
            if caffeinated:
                ctypes.windll.kernel32.SetThreadExecutionState(  # type: ignore[attr-defined]
                    _ES_CONTINUOUS | _ES_SYSTEM_REQUIRED)
            else:
                ctypes.windll.kernel32.SetThreadExecutionState(  # type: ignore[attr-defined]
                    _ES_CONTINUOUS)
        else:
            # TODO: Support on Linux.
            #       See the wakepy library for an example implementation.
            # 
            # Not supported
            return
        cls._caffeinated = caffeinated



"""
Tests that projects in older on-disk formats are upgraded to the latest
format properly.

In particular this module tests the Project._apply_migrations() method.

All tests below implicitly include the condition:
* given_project_opened_as_writable
"""
from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager, contextmanager
from crystal import progress
import crystal.main
from crystal.model import Project, Resource
from crystal.model import ResourceRevision as RR
from crystal.model import ResourceRevisionMetadata
from crystal.progress import CancelOpenProject, OpenProjectProgressDialog
from crystal.tests.test_server import serve_and_fetch_xkcd_home_page
from crystal.tests.util.controls import click_button
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import extracted_project
from crystal.tests.util.wait import wait_for, wait_while, window_condition
from crystal.tests.util.windows import MainWindow, OpenOrCreateDialog
from crystal.util.db import DatabaseCursor
from crystal.util.wx_dialog import mocked_show_modal
import os.path
from typing_extensions import override
from unittest import skip
from unittest.mock import patch
import wx

# === .crystalopen & README ===

@skip('not yet automated')
def test_given_project_lacks_crystalopen_and_lacks_readme_when_project_opened_then_crystalopen_and_readme_created() -> None:
    pass


@skip('not yet automated')
def test_given_project_lacks_crystalopen_and_has_maybe_modified_readme_when_project_opened_then_readme_content_preserved() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_crystalopen_with_nondefault_name_when_project_opened_then_crystalopen_name_preserved() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_crystalopen_and_has_maybe_modified_readme_when_project_opened_then_readme_content_preserved() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_crystalopen_and_readme_was_deleted_when_project_opened_then_readme_stays_deleted() -> None:
    pass


# === Windows desktop.ini ===

@skip('not yet automated')
def test_given_project_lacks_desktop_ini_file_then_desktop_ini_and_icons_directory_created() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_desktop_ini_file_then_desktop_ini_content_preserved() -> None:
    pass


# === Linux .directory ===

@skip('not yet automated')
def test_given_project_lacks_dot_directory_file_then_dot_directory_file_created() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_dot_directory_file_then_dot_directory_file_content_preserved() -> None:
    pass


# === Hide special files on Windows ===

@skip('not yet automated')
def test_given_windows_then_desktop_ini_file_and_dot_directory_file_marked_as_hidden() -> None:
    pass


# === Major Version Migrations ===

# --- Unknown Major Versions ---

async def test_refuses_to_open_project_with_unknown_high_major_version() -> None:
    UNKNOWN_HIGH_MAJOR_VERSION = Project._LATEST_SUPPORTED_MAJOR_VERSION + 1
    
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with unknown high major version
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            project._set_property('major_version', str(UNKNOWN_HIGH_MAJOR_VERSION))
            assert UNKNOWN_HIGH_MAJOR_VERSION == project.major_version
        
        # Try to open that project
        if True:
            # Prepare to: OK
            did_respond_to_project_too_new_modal = False
            def click_ok_in_project_too_new_modal(dialog: wx.Dialog) -> int:
                assert 'cr-project-too-new' == dialog.Name
                
                nonlocal did_respond_to_project_too_new_modal
                did_respond_to_project_too_new_modal = True
                
                return wx.ID_OK
            
            load_project = crystal.main._load_project  # capture
            def patched_load_project(*args, **kwargs) -> Project:
                return load_project(
                    *args,
                    _show_modal_func=click_ok_in_project_too_new_modal,  # type: ignore[misc]
                    **kwargs)
            
            with patch(f'crystal.main._load_project', patched_load_project):
                ocd = await OpenOrCreateDialog.wait_for()
                await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
                
                # HACK: Wait minimum duration to allow open to finish
                await bg_sleep(0.5)
                
                # Wait for cancel and return to initial dialog
                ocd = await OpenOrCreateDialog.wait_for()
                
                assert did_respond_to_project_too_new_modal


# --- 1 -> 2: Prompt to upgrade ---

@skip('covered by: test_can_upgrade_project_from_major_version_1_to_2')
async def test_given_project_is_major_version_1_and_would_be_fast_to_upgrade_to_major_version_2_when_open_project_then_updates_project_without_prompting_for_confirmation() -> None:
    pass


async def test_when_prompted_to_upgrade_project_from_major_version_1_to_2_then_can_continue_upgrade() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        def continue_button_id(dialog: wx.Dialog) -> int:
            style = dialog.GetMessageDialogStyle()
            if style & wx.YES_NO != 0:
                return wx.ID_YES  # Continue
            elif style & wx.OK != 0:
                return wx.ID_OK  # Continue
            else:
                raise AssertionError()
        
        with _upgrade_required_modal_always_shown(), \
                patch(
                    'crystal.progress.ShowModal',
                    # Prepare to: Accept upgrade
                    mocked_show_modal('cr-upgrade-required', continue_button_id)
                    ) as show_modal_method:
            async with (await OpenOrCreateDialog.wait_for()).open(
                    project_dirpath, wait_func=_wait_for_project_to_upgrade) as (mw, project):
                assert 1 == show_modal_method.call_count
                
                assert project.major_version >= 2, (
                    'Expected project to be upgraded'
                )  # not == 1


async def test_when_prompted_to_upgrade_project_from_major_version_1_to_2_then_can_defer_upgrade_to_later() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        def later_button_id(dialog: wx.Dialog) -> int:
            style = dialog.GetMessageDialogStyle()
            if style & wx.YES_NO != 0:
                return wx.ID_NO  # Later
            elif style & wx.OK != 0:
                raise AssertionError('"Later" button in dialog is not visible')
            else:
                raise AssertionError()
        
        with _upgrade_required_modal_always_shown(), \
                patch(
                    'crystal.progress.ShowModal',
                    # Prepare to: Defer upgrade
                    mocked_show_modal('cr-upgrade-required', later_button_id)
                    ) as show_modal_method:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                assert 1 == show_modal_method.call_count
                
                assert 1 == project.major_version, (
                    'Expected project to not be upgraded'
                )  # not > 1


async def test_when_prompted_to_upgrade_project_from_major_version_1_to_2_then_can_cancel_open_project() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        with _upgrade_required_modal_always_shown(), \
                patch(
                    'crystal.progress.ShowModal',
                    # Prepare to: Cancel
                    mocked_show_modal('cr-upgrade-required', wx.ID_CANCEL)
                    ) as show_modal_method:
            ocd = await OpenOrCreateDialog.wait_for()
            await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
            
            # HACK: Wait minimum duration to allow open to finish
            await bg_sleep(0.5)
            
            # Wait for cancel and return to initial dialog
            ocd = await OpenOrCreateDialog.wait_for()
            assert 1 == show_modal_method.call_count


async def test_given_project_is_major_version_1_and_has_more_than_MAX_REVISION_ID_revisions_when_open_project_then_defers_upgrade_to_later() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        # Open project, simulating an estimated revision count > Project._MAX_REVISION_ID.
        # Ensure is not upgraded to major version 2.
        if True:
            super_process_table_rows = Project._process_table_rows
            @staticmethod  # type: ignore[misc]
            def process_table_rows(
                    c: DatabaseCursor,
                    approx_row_count_query: str,
                    *args, **kwargs,
                    ) -> None:
                return super_process_table_rows(
                    c,
                    f'select {Project._MAX_REVISION_ID + 1}'
                        if approx_row_count_query.startswith('select id from resource_revision ')
                        else approx_row_count_query,
                    *args, **kwargs)
        
        with patch.object(Project, '_process_table_rows', process_table_rows):
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                assert 1 == project.major_version, (
                    'Expected project to not be upgraded'
                )  # not > 1


# --- 1 -> 2: Move revisions from flat structure to hierarchy ---

async def test_can_upgrade_project_from_major_version_1_to_2() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with >=4,096 revisions, which is just over the number of
        # revisions that fit into a single subdirectory of the new hierarchy
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
            
            # Create many resources
            max_id = max([r._id for r in project.resources])
            resources = Resource.bulk_create(project, [
                f'https://example.com/{id}'
                for id in range(max_id + 1, 4096 + 1)
            ], 'https://example.com/')
            max_id = max([r._id for r in project.resources])
            assert 4096 == max_id
            
            # Create many revisions
            if True:
                # Create revision rows in database
                c = project._db.cursor()
                encoded_null_error = RR._encode_error(None)
                encoded_404_error = RR._encode_metadata(ResourceRevisionMetadata(
                    http_version=10,
                    status_code=404,
                    reason_phrase='Not Found',
                    headers=[],
                ))
                c.executemany(
                    'insert into resource_revision '
                        '(id, resource_id, request_cookie, error, metadata) values (?, ?, ?, ?, ?)', 
                    [
                        (r._id, r._id, None, encoded_null_error, encoded_404_error)
                        for r in resources
                    ])
                project._db.commit()
                
                # Create revision files in filesystem
                for id in [r._id for r in resources]:
                    with open(os.path.join(project.path, Project._REVISIONS_DIRNAME, str(id)), 'w') as tf:
                        tf.write(str(id))  # arbitrary content
        
        # Upgrade the project to major version >= 2.
        # Ensure revisions appear to be migrated correctly.
        async with (await OpenOrCreateDialog.wait_for()).open(
                project_dirpath, wait_func=_wait_for_project_to_upgrade) as (mw, project):
            assert project.major_version >= 2, (
                'Expected project to be upgraded'
            )  # not == 1
            
            assert os.path.isdir(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '000'))
            assert os.path.isfile(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '000', '001'))
            
            assert os.path.isdir(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '001'))
            assert os.path.isfile(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '001', '000'))
            assert not os.path.exists(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '001', '001'))
            
            assert not os.path.exists(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '002'))


async def test_can_cancel_and_resume_upgrade_of_project_from_major_version_1_to_2() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        old_revision_count = _count_files_in(
            os.path.join(project.path, Project._REVISIONS_DIRNAME))  # capture
        assert old_revision_count >= 1
        
        # Start upgrading the project to major version >= 2,
        # but cancel the upgrade after the second revision
        if True:
            ocd = await OpenOrCreateDialog.wait_for()
            
            progress_listener = progress._active_progress_listener
            assert progress_listener is not None
            
            # Prepare to: Cancel upgrade in the middle
            did_cancel_migration_in_middle = False  # type: bool
            def upgrading_revision(index: int, revisions_per_second: float) -> None:
                nonlocal did_cancel_migration_in_middle
                if index >= 1:
                    # Cancel before migrating the second revision
                    did_cancel_migration_in_middle = True
                    raise CancelOpenProject()
            
            with patch.object(progress_listener, 'upgrading_revision', upgrading_revision), \
                    _progress_reported_at_maximum_resolution():
                await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
                
                # HACK: Wait minimum duration to allow open to finish
                await bg_sleep(0.5)
                
                # Wait for migration to start, get cancelled, and return to initial dialog
                ocd = await OpenOrCreateDialog.wait_for()
                assert did_cancel_migration_in_middle
        
        # Verify looks like migration was canceled in filesystem
        assert os.path.exists(os.path.join(project.path, Project._REVISIONS_DIRNAME))
        assert os.path.exists(os.path.join(project.path, Project._IN_PROGRESS_REVISIONS_DIRNAME))
        assert 1 == _count_files_in(
            os.path.join(project.path, Project._IN_PROGRESS_REVISIONS_DIRNAME))
        assert (old_revision_count - 1) == _count_files_in(
            os.path.join(project.path, Project._REVISIONS_DIRNAME))
        
        # Resume upgrading the project to major version >= 2. Allow to finish.
        if True:
            async with (await OpenOrCreateDialog.wait_for()).open(
                    project_dirpath, wait_func=_wait_for_project_to_upgrade) as (mw, project):
                assert project.major_version >= 2, (
                    'Expected project to be upgraded'
                )  # not == 1
            new_revision_count = _count_files_in(
                os.path.join(project.path, Project._REVISIONS_DIRNAME))
        
        # Ensure all revisions appear to be migrated
        assert os.path.exists(os.path.join(project.path, Project._REVISIONS_DIRNAME))
        assert not os.path.exists(os.path.join(project.path, Project._IN_PROGRESS_REVISIONS_DIRNAME))
        assert old_revision_count == new_revision_count
        assert not os.path.exists(os.path.join(project.path, Project._TEMPORARY_DIRNAME, Project._REVISIONS_DIRNAME))


# --- 1 -> 2: Smoke test operations on both project versions ---

async def test_can_serve_revisions_from_project_with_major_version_1() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        async with _project_opened_without_migrating(project_dirpath) as (mw, project):
            assert 1 == project.major_version
            
            (server_page, _) = await serve_and_fetch_xkcd_home_page(mw)
            assert 200 == server_page.status


async def test_can_serve_revisions_from_project_with_major_version_2() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            assert 2 == project.major_version
            
            (server_page, _) = await serve_and_fetch_xkcd_home_page(mw)
            assert 200 == server_page.status


# --- Utility ---

@asynccontextmanager
async def _project_opened_without_migrating(
        project_dirpath: str
        ) -> AsyncIterator[tuple[MainWindow, Project]]:
    class NonUpgradingProject(Project):
        @override
        def _apply_migrations(self, *args, **kwargs) -> None:
            pass
    
    # Project imported inside by _prompt_to_open_project()
    with patch('crystal.model.Project', NonUpgradingProject), \
            patch('crystal.tests.util.windows.Project', NonUpgradingProject):
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
            yield (mw, project)


@contextmanager
def _upgrade_required_modal_always_shown() -> Iterator[None]:
    was_enabled = OpenProjectProgressDialog._always_show_upgrade_required_modal
    OpenProjectProgressDialog._always_show_upgrade_required_modal = True
    try:
        yield
    finally:
        OpenProjectProgressDialog._always_show_upgrade_required_modal = was_enabled


@contextmanager
def _progress_reported_at_maximum_resolution() -> Iterator[None]:
    was_enabled = Project._report_progress_at_maximum_resolution
    Project._report_progress_at_maximum_resolution = True
    try:
        yield
    finally:
        Project._report_progress_at_maximum_resolution = was_enabled


def _count_files_in(dirpath: str) -> int:
    file_count = 0
    for (_, _, filenames) in os.walk(dirpath):
        file_count += len(filenames)
    return file_count


async def _wait_for_project_to_upgrade() -> None:
    if OpenProjectProgressDialog._upgrading_revision_progress is None:
        OpenProjectProgressDialog._upgrading_revision_progress = 0
    
    def progression_func() -> int | None:
        return OpenProjectProgressDialog._upgrading_revision_progress
    await wait_while(progression_func)



from urllib.parse import quote


# Adapted from requote_uri() @ requests/utils.py
def requote_uri(uri):
    # type: (str) -> str
    """
    Re-quote the given URI.

    This function passes the given URI through an unquote/quote cycle to
    ensure that it is fully and consistently quoted.
    """
    safe_with_percent = "!#$%&'()*+,/:;=?@[]~"
    safe_without_percent = "!#$&'()*+,/:;=?@[]~"
    try:
        # Unquote only the unreserved characters
        # Then quote only illegal characters (do not quote reserved,
        # unreserved, or '%')
        return quote(_unquote_unreserved(uri), safe=safe_with_percent)
    except _InvalidURL:
        # We couldn't unquote the given URI, so let's try quoting it, but
        # there may be unquoted '%'s in the URI. We need to make sure they're
        # properly quoted so they do not cause issues elsewhere.
        return quote(uri, safe=safe_without_percent)


# The unreserved URI characters (RFC 3986)
_UNRESERVED_SET = frozenset(
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz" + "0123456789-._~")

# Adapted from unquote_unreserved() @ requests/utils.py
def _unquote_unreserved(uri):
    # type: (str) -> str
    """
    Un-escape any percent-escape sequences in a URI that are unreserved
    characters. This leaves all reserved, illegal and non-ASCII bytes encoded.
    """
    parts = uri.split('%')
    for i in range(1, len(parts)):
        h = parts[i][0:2]
        if len(h) == 2 and h.isalnum():
            try:
                c = chr(int(h, 16))
            except ValueError:
                raise _InvalidURL("Invalid percent-escape sequence: '%s'" % h)

            if c in _UNRESERVED_SET:
                parts[i] = c + parts[i][2:]
            else:
                parts[i] = '%' + parts[i]
        else:
            parts[i] = '%' + parts[i]
    return ''.join(parts)


class _InvalidURL(ValueError):
    """The URL provided was somehow invalid."""


def is_unrewritable_url(relative_url: str) -> bool:
    # Don't rewrite certain schemes
    for prefix in ('mailto:', 'javascript:', 'data:'):
        if relative_url.startswith(prefix):
            return True
    
    # Otherwise OK to rewrite
    return False



from collections.abc import Callable, Iterator
from contextlib import contextmanager
import os
from typing import NoReturn

# Whether RuntimeError('wrapped C/C++ object of type ... has been deleted')
# (or its sibling WindowDeletedError) is silently ignored.
# 
# It is unsafe to ignore these errors in general because they are raised
# on only a best effort basis when a use-after-free error in C/C++ would
# normally have happened and either crashed the process immediately
# or corrupted memory, crashing the process later.
IGNORE_USE_AFTER_FREE = (
    os.environ.get('CRYSTAL_IGNORE_USE_AFTER_FREE', 'False') == 'True'
)


def is_wrapped_object_deleted_error(e: Exception) -> bool:
    e_str = str(e)
    return (
        isinstance(e, RuntimeError) and
        e_str.startswith('wrapped C/C++ object of type ') and
        e_str.endswith(' has been deleted')
    )


@contextmanager
def wrapped_object_deleted_error_ignored() -> Iterator[None]:
    if IGNORE_USE_AFTER_FREE:
        try:
            yield
        except Exception as e:
            if is_wrapped_object_deleted_error(e):
                pass
            else:
                raise
    else:
        # May raise errors matching is_wrapped_object_deleted_error()
        yield


@contextmanager
def wrapped_object_deleted_error_raising(
        raiser: Callable[[], NoReturn] | None=None
        ) -> Iterator[None]:
    if IGNORE_USE_AFTER_FREE:
        if raiser is None:
            def default_raiser() -> NoReturn:
                raise WindowDeletedError()
            raiser = default_raiser
        try:
            yield
        except Exception as e:
            if is_wrapped_object_deleted_error(e):
                raiser()
            else:
                raise
    else:
        # May raise errors matching is_wrapped_object_deleted_error()
        yield
            

class WindowDeletedError(Exception):
    """Raises when code attempts to manipulate a wx.Window that has been deleted."""



from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager, contextmanager
from crystal.model import Project, Resource, ResourceGroup, RootResource
from crystal.task import (
    DownloadResourceGroupMembersTask, DownloadResourceGroupTask,
    DownloadResourceTask,
)
from crystal.tests.util.downloads import load_children_of_drg_task
from crystal.tests.util.server import served_project
from crystal.tests.util.tasks import (
    append_deferred_top_level_tasks, clear_top_level_tasks_on_exit,
    scheduler_disabled, step_scheduler, step_scheduler_until_done,
)
from crystal.tests.util.wait import wait_for
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.util.wx_dialog import mocked_show_modal
from unittest.mock import patch
import wx


async def test_when_reopen_project_given_resource_group_was_downloading_then_resumes_downloading() -> None:
    # ...starting from correct member
    
    with scheduler_disabled(), served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        comic1_url = sp.get_request_url('https://xkcd.com/1/')
        comic2_url = sp.get_request_url('https://xkcd.com/2/')
        comic3_url = sp.get_request_url('https://xkcd.com/3/')
        comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        # Create new project, with a partially downloaded group
        async with (await OpenOrCreateDialog.wait_for()).create(delete=False) as (mw, project):
            project_dirpath = project.path
            
            comic1_rr = RootResource(project, '', Resource(project, comic1_url))
            comic2_rr = RootResource(project, '', Resource(project, comic2_url))
            comic3_rr = RootResource(project, '', Resource(project, comic3_url))
            
            comic_g = ResourceGroup(project, '', comic_pattern, source=None)
            comic_g.download()
            
            # Ensure task list has only a task to download the group
            append_deferred_top_level_tasks(project)
            (drg_task,) = project.root_task.children
            assert isinstance(drg_task, DownloadResourceGroupTask)
            assert comic_g == drg_task.group
            
            # Step scheduler until comic #1 and #2 are fully downloaded
            # (including embedded resources), but not #3
            if True:
                load_children_of_drg_task(drg_task, scheduler_thread_enabled=False)
                (_, drgm_task) = drg_task.children
                assert isinstance(drgm_task, DownloadResourceGroupMembersTask)
                
                (comic1_dr_task, comic2_dr_task, comic3_dr_task, *_) = drgm_task.children
                assert isinstance(comic1_dr_task, DownloadResourceTask)
                assert isinstance(comic2_dr_task, DownloadResourceTask)
                assert isinstance(comic3_dr_task, DownloadResourceTask)
                
                while not (comic1_dr_task.complete and 
                        comic2_dr_task.complete):
                    await step_scheduler(project)
                assert not comic3_dr_task.complete
                
                assert comic1_rr.resource.has_any_revisions()
                assert comic2_rr.resource.has_any_revisions()
                assert not comic3_rr.resource.has_any_revisions()
            
            # (Close the project, with top-level tasks still running)
        
        # Reopen same project, and resume downloads
        async with _open_project_with_resume_data(project_dirpath, resume=True) as project:
            assert project is not None
            with clear_top_level_tasks_on_exit(project):
                # Ensure comic #1 and #2 are downloaded, but not #3
                if True:
                    (comic1_rr, comic2_rr, comic3_rr) = project.root_resources
                    
                    assert comic1_rr.resource.has_any_revisions()
                    assert comic2_rr.resource.has_any_revisions()
                    assert not comic3_rr.resource.has_any_revisions()
                    
                    assert comic1_rr.resource.already_downloaded_this_session
                    assert comic2_rr.resource.already_downloaded_this_session
                    assert not comic3_rr.resource.already_downloaded_this_session
                
                (comic_g,) = project.resource_groups
                
                # Ensure task list has only a task to download the group
                append_deferred_top_level_tasks(project)
                (drg_task,) = project.root_task.children
                assert isinstance(drg_task, DownloadResourceGroupTask)
                assert comic_g == drg_task.group
                
                # Ensure group download task has children showing that
                # comic #1 and #2 are downloaded, but not #3
                if True:
                    load_children_of_drg_task(drg_task, scheduler_thread_enabled=False)
                    (_, drgm_task) = drg_task.children
                    assert isinstance(drgm_task, DownloadResourceGroupMembersTask)
                    
                    (comic1_dr_task, comic2_dr_task, comic3_dr_task, *_) = drgm_task.children
                    assert isinstance(comic1_dr_task, DownloadResourceTask)
                    assert isinstance(comic2_dr_task, DownloadResourceTask)
                    assert isinstance(comic3_dr_task, DownloadResourceTask)
                    assert comic1_dr_task.complete
                    assert comic2_dr_task.complete
                    assert not comic3_dr_task.complete
                    
                    def assert_first_executed_child_is_correct() -> None:
                        assert (
                            drgm_task.children.index(comic3_dr_task) ==
                            drgm_task._next_child_index
                        )
                    await step_scheduler(project, after_get=assert_first_executed_child_is_correct)
                
                # Step scheduler until comic #3 downloaded
                while not comic3_dr_task.complete:
                    await step_scheduler(project)


async def test_when_reopen_project_given_resource_was_downloading_then_resumes_downloading() -> None:
    with scheduler_disabled(), served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        comic1_url = sp.get_request_url('https://xkcd.com/1/')
        
        # Create new project, with a partially downloaded list of resources
        async with (await OpenOrCreateDialog.wait_for()).create(delete=False) as (mw, project):
            project_dirpath = project.path
            
            atom_feed_rr = RootResource(project, '', Resource(project, atom_feed_url))
            comic1_rr = RootResource(project, '', Resource(project, comic1_url))
            
            # Start downloading resources
            atom_feed_rr.download()
            comic1_rr.download()
            
            append_deferred_top_level_tasks(project)
            (atom_feed_dr_task, comic1_dr_task) = project.root_task.children
            assert isinstance(atom_feed_dr_task, DownloadResourceTask)
            assert isinstance(comic1_dr_task, DownloadResourceTask)

            # Step scheduler until resource #1 downloaded
            while not atom_feed_dr_task.complete:
                await step_scheduler(project)
            assert not comic1_dr_task.complete
        
        # Reopen same project, and resume downloads
        async with _open_project_with_resume_data(project_dirpath, resume=True) as project:
            with clear_top_level_tasks_on_exit(project):
                (atom_feed_rr, comic1_rr) = project.root_resources
                
                # Ensure task list has only tasks to download the remaining resources
                append_deferred_top_level_tasks(project)
                (comic1_dr_task,) = project.root_task.children
                assert isinstance(comic1_dr_task, DownloadResourceTask)
                assert comic1_rr.resource == comic1_dr_task.resource
                
                # Step scheduler until resource #2 downloaded
                while not comic1_dr_task.complete:
                    await step_scheduler(project)
                await step_scheduler(project, expect_done=True)
                
                # Ensure no download tasks left
                () = project.root_task.children


async def test_when_reopen_project_and_user_cancels_resume_then_does_not_resume_downloading() -> None:
    with scheduler_disabled(), served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        comic1_url = sp.get_request_url('https://xkcd.com/1/')

        # Create new project, with a partially downloaded list of resources
        async with (await OpenOrCreateDialog.wait_for()).create(delete=False) as (mw, project):
            project_dirpath = project.path
            
            atom_feed_rr = RootResource(project, '', Resource(project, atom_feed_url))
            comic1_rr = RootResource(project, '', Resource(project, comic1_url))
            
            atom_feed_rr.download()
            comic1_rr.download()
            
            append_deferred_top_level_tasks(project)
            (atom_feed_dr_task, comic1_dr_task) = project.root_task.children
            assert isinstance(atom_feed_dr_task, DownloadResourceTask)
            assert isinstance(comic1_dr_task, DownloadResourceTask)
            
            # Step scheduler until resource #1 downloaded
            while not atom_feed_dr_task.complete:
                await step_scheduler(project)
            assert not comic1_dr_task.complete
        
        # Reopen same project, and cancel resume
        async with _open_project_with_resume_data(project_dirpath, resume=False) as project:
            # Ensure no download tasks are resumed
            append_deferred_top_level_tasks(project)
            () = project.root_task.children


async def test_when_reopen_project_as_readonly_then_does_not_resume_downloading() -> None:
    with scheduler_disabled(), served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        comic1_url = sp.get_request_url('https://xkcd.com/1/')

        # Create new project, with a partially downloaded list of resources
        async with (await OpenOrCreateDialog.wait_for()).create(delete=False) as (mw, project):
            project_dirpath = project.path
            
            atom_feed_rr = RootResource(project, '', Resource(project, atom_feed_url))
            comic1_rr = RootResource(project, '', Resource(project, comic1_url))
            
            atom_feed_rr.download()
            comic1_rr.download()
            
            append_deferred_top_level_tasks(project)
            (atom_feed_dr_task, comic1_dr_task) = project.root_task.children
            assert isinstance(atom_feed_dr_task, DownloadResourceTask)
            assert isinstance(comic1_dr_task, DownloadResourceTask)
            
            # Step scheduler until resource #1 downloaded
            while not atom_feed_dr_task.complete:
                await step_scheduler(project)
            assert not comic1_dr_task.complete
        
        # Reopen same project as read-only
        async with (await OpenOrCreateDialog.wait_for()).open(
                project_dirpath, readonly=True) as (mw, project):
            # Ensure no download tasks are resumed
            append_deferred_top_level_tasks(project)
            () = project.root_task.children
        
        # Reopen same project normally, and resume downloads
        async with _open_project_with_resume_data(project_dirpath, resume=True) as project:
            with clear_top_level_tasks_on_exit(project):
                # Ensure download tasks are resumed
                append_deferred_top_level_tasks(project)
                (comic1_dr_task,) = project.root_task.children
                assert isinstance(comic1_dr_task, DownloadResourceTask)


async def test_when_close_project_abruptly_and_reopen_project_with_stale_resume_data_then_resume_ignores_invalid_data() -> None:
    with scheduler_disabled(), served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        comic1_url = sp.get_request_url('https://xkcd.com/1/')
        comic2_url = sp.get_request_url('https://xkcd.com/2/')
        feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        
        # Create a project with stale resume data referring to a deleted ResourceGroup and Resource
        async with (await OpenOrCreateDialog.wait_for()).create(delete=False) as (mw, project):
            project_dirpath = project.path
            
            comic_g = ResourceGroup(project, '', comic_pattern, source=None)
            RootResource(project, '', Resource(project, comic1_url))
            comic2_rr = RootResource(project, '', Resource(project, comic2_url))
            feed_rr = RootResource(project, '', Resource(project, feed_url))
            
            # Start downloading a ResourceGroup and a Resource
            if True:
                comic_g.download()
                feed_rr.download()
                comic2_rr.download()  # will NOT delete later
                
                append_deferred_top_level_tasks(project)
                (drg_task, dr_task1, dr_task2) = project.root_task.children
                assert isinstance(drg_task, DownloadResourceGroupTask)
                assert isinstance(dr_task1, DownloadResourceTask)
                assert isinstance(dr_task2, DownloadResourceTask)
                assert comic_g == drg_task.group
                assert feed_rr.resource == dr_task1.resource
                assert comic2_rr.resource == dr_task2.resource
            
            # Create resume data referring to the ResourceGroup and Resource
            _autohibernate_now(project)
            
            # Forget the ResourceGroup and Resource. Ensure both continue to download anyway.
            if True:
                comic_g.delete()
                feed_rr.delete()
                feed_rr.resource.delete()  # HACK: Unrealistic. No way to delete Resource in the UI.
                # (comic2_rr is not deleted so it WILL be resumed later)
                
                append_deferred_top_level_tasks(project)
                (drg_task, dr_task1, dr_task2) = project.root_task.children
                assert isinstance(drg_task, DownloadResourceGroupTask)
                assert isinstance(dr_task1, DownloadResourceTask)
                assert isinstance(dr_task2, DownloadResourceTask)
                assert comic_g == drg_task.group
                assert feed_rr.resource == dr_task1.resource
                assert comic2_rr.resource == dr_task2.resource
            
            _close_project_abruptly(project)
        
        # Reopen the project
        async with _open_project_with_resume_data(project_dirpath, resume=True) as project:
            with clear_top_level_tasks_on_exit(project):
                # Ensure that the stale resume data is ignored,
                # and the download of the non-stale Resource is resumed
                append_deferred_top_level_tasks(project)
                (dr_task2,) = project.root_task.children
                assert isinstance(dr_task2, DownloadResourceTask)


async def test_when_close_project_normally_and_no_tasks_running_then_resume_data_in_database_is_cleared() -> None:
    with scheduler_disabled(), served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        comic1_url = sp.get_request_url('https://xkcd.com/1/')
        
        # Create new project, but with no downloads running at the time of closing
        async with (await OpenOrCreateDialog.wait_for()).create(delete=False) as (mw, project):
            project_dirpath = project.path
            
            comic1_rr = RootResource(project, '', Resource(project, comic1_url))
            comic1_rr.download()
            append_deferred_top_level_tasks(project)
            (comic1_dr_task,) = project.root_task.children
            assert isinstance(comic1_dr_task, DownloadResourceTask)
            
            _autohibernate_now(project)
            assert _project_has_resume_data(project)
            
            await step_scheduler_until_done(project)

        # Ensure resume data is cleared
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):

            assert not _project_has_resume_data(project)


async def test_while_project_open_then_periodically_saves_resume_data_so_that_can_resume_after_abrupt_project_close() -> None:
    with scheduler_disabled(), served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        atom_feed_url = sp.get_request_url('https://xkcd.com/atom.xml')
        comic1_url = sp.get_request_url('https://xkcd.com/1/')

        with _frequent_autohibernates_enabled():
            # Create new project, with a partially downloaded list of resources
            async with (await OpenOrCreateDialog.wait_for()).create(delete=False) as (mw, project):
                project_dirpath = project.path
                
                atom_feed_rr = RootResource(project, '', Resource(project, atom_feed_url))
                comic1_rr = RootResource(project, '', Resource(project, comic1_url))
                
                atom_feed_rr.download()
                comic1_rr.download()
                
                append_deferred_top_level_tasks(project)
                (atom_feed_dr_task, comic1_dr_task) = project.root_task.children
                assert isinstance(atom_feed_dr_task, DownloadResourceTask)
                assert isinstance(comic1_dr_task, DownloadResourceTask)
                
                # Step scheduler until resource #1 downloaded
                while not atom_feed_dr_task.complete:
                    await step_scheduler(project)
                assert not comic1_dr_task.complete
                
                # Wait for autohibernate to save resume data
                _clear_project_resume_data(project)
                assert not _project_has_resume_data(project)
                await wait_for(lambda: _project_has_resume_data(project) or None, timeout=2)
                
                _close_project_abruptly(project)
            
            # Reopen the project
            async with _open_project_with_resume_data(project_dirpath, resume=True) as project:
                with clear_top_level_tasks_on_exit(project):
                    # Ensure that the download is resumed
                    append_deferred_top_level_tasks(project)
                    (comic1_dr_task,) = project.root_task.children
                    assert isinstance(comic1_dr_task, DownloadResourceTask)


# === Utility ===

@asynccontextmanager
async def _open_project_with_resume_data(
        project_dirpath: str,
        resume: bool=True,
        ) -> AsyncIterator[Project]:
    """
    Opens a project with resume data, simulating the behavior of the
    `OpenOrCreateDialog` when it asks the user whether to resume downloads.
    
    If `resume` is `True`, it simulates the user clicking "Resume Downloads".
    If `resume` is `False`, it simulates the user clicking "Cancel".
    """
    with patch(
            'crystal.browser.ShowModal',
            mocked_show_modal('cr-resume-downloads', wx.ID_OK if resume else wx.ID_CANCEL)
            ) as show_modal_method:
        async def wait_for_project_to_unhibernate() -> None:
            await wait_for(lambda: (1 == show_modal_method.call_count) or None)
        
        async with (await OpenOrCreateDialog.wait_for()).open(
                project_dirpath, wait_func=wait_for_project_to_unhibernate) as (mw, project):
            yield project


def _autohibernate_now(project: Project) -> None:
    """
    Simulates the automatic hibernation of a project.
    """
    project.hibernate_tasks()


@contextmanager
def _frequent_autohibernates_enabled() -> Iterator[None]:
    """
    Enables frequent automatic hibernation for projects that become opened.
    """
    with patch(
            'crystal.browser.MainWindow._AUTOHIBERNATE_PERIOD',
            20  # milliseconds
            ):
        yield


def _close_project_abruptly(project: Project) -> None:
    """
    Simulates an abrupt close of a project, similar to unmounting the disk
    where the project database is stored.
    """
    # Causes subsequent SQLite operations to fail with:
    #     sqlite3.ProgrammingError('Cannot operate on a closed database.')
    # 
    # A real unmount of a disk on macOS causes subsequent SQLite operations to fail with:
    #     sqlite3.DatabaseError('database disk image is malformed')
    project._db._db.close()


def _project_has_resume_data(project: Project) -> bool:
    """
    Checks if the project has any resume data in the database.
    """
    hibernated_project_str = project._get_property('hibernated_state', default=None)
    return hibernated_project_str is not None


def _clear_project_resume_data(project: Project) -> None:
    """
    Clears the resume data from the project database.
    """
    project._delete_property('hibernated_state')



from collections.abc import Callable, Generator
from crystal.download import HTTP_REQUEST_TIMEOUT
from crystal.util.bulkheads import capture_crashes_to_stderr
from crystal.util.test_mode import tests_are_running
from crystal.util.xthreading import (
    bg_affinity, fg_affinity, start_thread_switching_coroutine, SwitchToThread,
)
import os
from typing import Tuple
from urllib.parse import urlparse, urlunparse
from urllib.request import urlopen


class UrlCleaner:
    """
    Given a potentially messy URL typed manually by the user,
    such as from a browser's URL bar, consider candidate URLs
    that the user might have meant, and resolve a single
    cleaned URL.
    """
    
    @fg_affinity
    def __init__(self,
            url_input: str,
            on_running_changed_func: Callable[[bool], None],
            set_cleaned_url_func: Callable[[str], None],
            ) -> None:
        self.url_input = url_input
        self._on_running_changed_func = on_running_changed_func
        self._set_cleaned_url_func = set_cleaned_url_func
        self._cancelled = False
    
    @fg_affinity
    def start(self) -> None:
        url_candidates = _candidate_urls_from_user_input(self.url_input)
        start_thread_switching_coroutine(
            SwitchToThread.FOREGROUND,
            self._run(url_candidates),
            capture_crashes_to_stderr,
        )
    
    @fg_affinity
    def _run(self, url_candidates: list[str]) -> Generator[SwitchToThread, None, None]:
        assert len(url_candidates) >= 1
        
        self._on_running_changed_func(True)
        if len(url_candidates) == 1:
            cleaned_url = url_candidates[0]
            self._finish(cleaned_url)
            return
        
        try:
            yield SwitchToThread.BACKGROUND
            try:
                cleaned_url = _resolve_url_from_candidates(
                    url_candidates, lambda: self._cancelled)
            except _ResolveCanceled:
                yield SwitchToThread.FOREGROUND
                assert self._cancelled
                return
            except:
                # Fallback to first candidate
                cleaned_url = url_candidates[0]
                raise
        finally:
            yield SwitchToThread.FOREGROUND
            if self._cancelled:
                return
            self._finish(cleaned_url)
    
    @fg_affinity
    def _finish(self, cleaned_url: str | None) -> None:
        if cleaned_url is not None:
            self._set_cleaned_url_func(cleaned_url)
        self._on_running_changed_func(False)
    
    # === Operations ===
    
    @fg_affinity
    def cancel(self) -> None:
        self._cancelled = True
        self._finish(None)


def cleaned_url_is_at_site_root(url_input: str) -> bool:
    candidates = _candidate_urls_from_user_input(url_input)
    assert len(candidates) >= 1
    candidate = candidates[0]
    return urlparse(candidate).path == '/'


def _candidate_urls_from_user_input(url_input: str) -> list[str]:
    """
    Given a potentially messy URL typed manually by the user,
    such as from a browser's URL bar, return candidate URLs
    that the user might have meant.
    
    Always at least one candidate is returned.
    
    In the current implementation:
    * If no https:// or http:// prefix is given,
      try first the former then the latter.
    * If additionally no www. domain prefix is given,
      try first without the prefix then with the prefix.
    """
    url_input = url_input.strip()  # remove leading and trailing whitespace
    
    if url_input == '':
        return [url_input]
    
    url_parts = urlparse(url_input)
    if url_parts.scheme in ('', 'https', 'http'):
        # If missing scheme:
        # 1. try https:// then http://
        # 2. try variations on www
        if url_parts.scheme == '':
            scheme_candidates = ('https', 'http')  # type: Tuple[str, ...]
            
            # Reparse (netloc='', path='DOMAIN/PATH')
            #      to (netloc='DOMAIN', path='PATH')
            if url_parts.netloc == '':
                path_parts = url_parts.path.split('/', 1)
                if len(path_parts) == 2:
                    (new_netloc, new_path) = path_parts
                else:
                    assert len(path_parts) == 1
                    (new_netloc, new_path) = (path_parts[0], '')
                url_parts = url_parts._replace(
                    netloc=new_netloc,
                    path=new_path
                )  # reinterpret
            
            # 1. If has www, then try non-www domain if www domain fails
            # 2. If missing www, then try it if non-www domain fails
            if url_parts.netloc.startswith('www.'):
                netloc_candidates = (url_parts.netloc, url_parts.netloc[len('www.'):])  # type: Tuple[str, ...]
            elif url_parts.netloc != '' and not url_parts.netloc.startswith('www.'):
                netloc_candidates = (url_parts.netloc, 'www.' + url_parts.netloc)
            else:
                netloc_candidates = (url_parts.netloc,)
        else:
            scheme_candidates = (url_parts.scheme,)
            netloc_candidates = (url_parts.netloc,)
        
        # If empty path then normalize to /
        if url_parts.path == '':
            path = '/'
        else:
            path = url_parts.path
        
        # Generate URL candidates
        url_candidates = []
        for scheme in scheme_candidates:
            for netloc in netloc_candidates:
                url_candidates.append(urlunparse(url_parts._replace(
                    scheme=scheme,
                    netloc=netloc,
                    path=path,
                    fragment='',
                )))
        
        return url_candidates
    else:
        # Non-HTTP scheme. Accept URL as-is.
        return [url_input]


@bg_affinity
def _resolve_url_from_candidates(
        url_candidates: list[str],
        did_cancel_func: Callable[[], bool]) -> str:
    """
    Given a list of candidate URLs, return the first URL that can be fetched
    successfully.
    
    Raises:
    * _ResolveCanceled
    """
    if len(url_candidates) == 0:
        raise ValueError()
    
    # If only one candidate, return it immediately
    if len(url_candidates) == 1:
        return url_candidates[0]
    
    # Disallow network requests while running tests,
    # unless CRYSTAL_URLOPEN_MOCKED=True
    if tests_are_running():
        if os.environ.get('CRYSTAL_URLOPEN_MOCKED', 'False') == 'True':
            # OK
            pass
        else:
            raise AssertionError(
                'Attempting to resolve URL candidates with real '
                'network requests while automated tests are running. '
                
                'Please either (1) enter URLs with http:// or https:// '
                'schema that does not need to be resolved, or (2) '
                'mock the urlopen() function with something like '
                '_urlopen_responding_with().'
            )
    
    # Look for URL candidate which can be fetched successfully
    for url_candidate in url_candidates:
        if did_cancel_func():
            raise _ResolveCanceled()
        try:
            print(f'Probing URL: {url_candidate}')
            with urlopen(url_candidate, timeout=HTTP_REQUEST_TIMEOUT) as f:
                if (f.getcode() // 100) != 2:  # not HTTP 2xx Success
                    continue
                
                # If redirected to a different candidate, return that candidate
                redirected_url = f.geturl()
                if redirected_url in url_candidates:
                    return redirected_url
                
                # Return this candidate (which might itself be a redirect)
                return url_candidate
        except Exception:
            # One of:
            # 1. Network down
            # 2. Domain does not exist
            # 3. IP did not respond
            # 4. IP refused connection on port that was tried
            continue
    
    # Fallback to first URL candidate (which wasn't fetched successfully)
    return url_candidates[0]


class _ResolveCanceled(Exception):
    pass



from crystal.model import (
    DownloadErrorDict, Project, Resource, ResourceGroup, RootResource,
)
from crystal.tests.util.asserts import assertEqual, assertIn
from crystal.tests.util.controls import TreeItem
from crystal.tests.util.downloads import network_down
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import served_project
from crystal.tests.util.tasks import wait_for_download_to_start_and_finish
from crystal.tests.util.wait import (
    DEFAULT_WAIT_PERIOD, first_child_of_tree_item_is_not_loading_condition,
    wait_for,
)
from crystal.tests.util.windows import (
    MainWindow, MenuitemMissingError, OpenOrCreateDialog,
)
import locale
import os
import tempfile
from unittest import skip
import wx

# ------------------------------------------------------------------------------
# Test: EntityTree: Default Domain/Directory

async def test_given_resource_node_whose_path_is_slash_when_set_default_url_domain_to_it_then_node_displays_only_path() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        rr = RootResource(project, '', Resource(project, 'https://neocities.org/'))
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rrn = root_ti.find_child(rr.resource.url, project.default_url_prefix)
        
        assert rrn.Text != '/'
        await mw.entity_tree.set_default_domain_to_entity_at_tree_item(rrn)
        assert rrn.Text == '/'
        
        # test_given_resource_node_whose_path_is_slash_and_default_url_domain_matches_it_when_clear_default_url_domain_then_node_displays_full_url
        await mw.entity_tree.clear_default_domain_from_entity_at_tree_item(rrn)
        assert rrn.Text.startswith('https://neocities.org/')
        
        # test_given_resource_node_whose_path_is_slash_then_cannot_set_default_url_prefix_to_it
        try:
            await mw.entity_tree.set_default_directory_to_entity_at_tree_item(rrn)
        except MenuitemMissingError:
            pass
        else:
            raise AssertionError('Did not expect resource to offer option to: set_default_url_prefix')


@skip('covered by: test_given_resource_node_whose_path_is_slash_when_set_default_url_domain_to_it_then_node_displays_only_path')
async def test_given_resource_node_whose_path_is_slash_then_cannot_set_default_url_prefix_to_it() -> None:
    pass


@skip('covered by: test_given_resource_node_whose_path_is_slash_when_set_default_url_domain_to_it_then_node_displays_only_path')
async def test_given_resource_node_whose_path_is_slash_and_default_url_domain_matches_it_when_clear_default_url_domain_then_node_displays_full_url() -> None:
    pass


async def test_given_resource_node_whose_path_is_more_than_slash_when_set_default_url_domain_to_it_then_node_displays_only_path() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        rr = RootResource(project, '', Resource(project, 'https://neocities.org/~distantskies/'))
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rrn = root_ti.find_child(rr.resource.url, project.default_url_prefix)
        
        assert rrn.Text != '/~distantskies/'
        await mw.entity_tree.set_default_domain_to_entity_at_tree_item(rrn)
        assert rrn.Text == '/~distantskies/'
        
        # test_given_resource_node_whose_path_is_more_than_slash_and_default_url_domain_matches_it_when_clear_default_url_domain_then_node_displays_full_url
        await mw.entity_tree.clear_default_domain_from_entity_at_tree_item(rrn)
        assert rrn.Text.startswith('https://neocities.org/~distantskies/')
        
        # test_given_resource_node_whose_path_is_more_than_slash_when_set_default_url_prefix_to_it_then_node_displays_only_slash
        await mw.entity_tree.set_default_directory_to_entity_at_tree_item(rrn)
        assert rrn.Text == '/'
        
        # test_given_resource_node_whose_path_is_more_than_slash_and_default_url_prefix_matches_it_when_clear_default_url_prefix_then_node_displays_full_url
        await mw.entity_tree.clear_default_directory_from_entity_at_tree_item(rrn)
        assert rrn.Text.startswith('https://neocities.org/~distantskies/')


@skip('covered by: test_given_resource_node_whose_path_is_more_than_slash_when_set_default_url_domain_to_it_then_node_displays_only_path')
async def test_given_resource_node_whose_path_is_more_than_slash_when_set_default_url_prefix_to_it_then_node_displays_only_slash() -> None:
    pass


@skip('covered by: test_given_resource_node_whose_path_is_more_than_slash_when_set_default_url_domain_to_it_then_node_displays_only_path')
async def test_given_resource_node_whose_path_is_more_than_slash_and_default_url_domain_matches_it_when_clear_default_url_domain_then_node_displays_full_url() -> None:
    pass


@skip('covered by: test_given_resource_node_whose_path_is_more_than_slash_when_set_default_url_domain_to_it_then_node_displays_only_path')
async def test_given_resource_node_whose_path_is_more_than_slash_and_default_url_prefix_matches_it_when_clear_default_url_prefix_then_node_displays_full_url() -> None:
    pass


@skip('not yet automated')
async def test_given_resource_group_node_whose_path_is_slash_wildcard_when_set_default_url_domain_to_it_then_node_displays_only_path_pattern() -> None:
    pass


@skip('not yet automated')
async def test_given_resource_group_node_whose_path_is_slash_wildcard_then_cannot_set_default_url_prefix_to_it() -> None:
    pass


@skip('not yet automated')
async def test_given_resource_group_node_whose_path_is_slash_wildcard_and_default_url_domain_matches_it_when_clear_default_url_domain_then_node_displays_full_url_pattern() -> None:
    pass


@skip('not yet automated')
async def test_given_resource_group_node_whose_path_is_more_than_slash_literal_when_set_default_url_domain_to_it_then_node_displays_only_path_pattern() -> None:
    pass


@skip('not yet automated')
async def test_given_resource_group_node_whose_path_is_more_than_slash_literal_when_set_default_url_prefix_to_it_then_node_displays_only_slash() -> None:
    pass


@skip('not yet automated')
async def test_given_resource_group_node_whose_path_is_more_than_slash_literal_and_default_url_domain_matches_it_when_clear_default_url_domain_then_node_displays_full_url_pattern() -> None:
    pass


@skip('not yet automated')
async def test_given_resource_group_node_whose_path_is_more_than_slash_literal_and_default_url_prefix_matches_it_when_clear_default_url_prefix_then_node_displays_full_url_pattern() -> None:
    pass


async def test_when_selected_entity_changes_and_top_level_entity_menu_opened_then_appropriate_change_url_prefix_actions_shown() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        
        # Case: No entities selected
        selected_ti = TreeItem.GetSelection(mw.entity_tree.window)
        assert (selected_ti is None) or (selected_ti == root_ti)
        cup_actions = _change_url_prefix_actions_in_top_level_menu(mw)
        assertEqual(
            [
                ('Set As Default Domain', False),
            ],
            [
                (mi.ItemLabelText, mi.Enabled) for mi in 
                cup_actions
            ]
        )
        
        rr = RootResource(project, '', Resource(project, 'https://neocities.org/'))
        rr2 = RootResource(project, '', Resource(project, 'https://neocities.org/~distantskies/'))
        
        rrn = root_ti.find_child(rr.resource.url, project.default_url_prefix)
        rrn2 = root_ti.find_child(rr2.resource.url, project.default_url_prefix)
        
        # Case: URL with path / selected
        rrn.SelectItem()
        cup_actions = _change_url_prefix_actions_in_top_level_menu(mw)
        assertEqual(
            [
                ('Set As Default Domain', True),
            ],
            [
                (mi.ItemLabelText, mi.Enabled) for mi in 
                cup_actions
            ]
        )
        
        # Case: URL with path / selected; matches domain
        _select_change_url_prefix_action(mw, cup_actions[0])
        cup_actions = _change_url_prefix_actions_in_top_level_menu(mw)
        assertEqual(
            [
                ('Clear Default Domain', True),
            ],
            [
                (mi.ItemLabelText, mi.Enabled) for mi in 
                cup_actions
            ]
        )
        
        # Clear Default URL Domain/Prefix
        _select_change_url_prefix_action(mw, cup_actions[0])
        cup_actions = _change_url_prefix_actions_in_top_level_menu(mw)
        assertEqual(
            [
                ('Set As Default Domain', True),
            ],
            [
                (mi.ItemLabelText, mi.Enabled) for mi in 
                cup_actions
            ]
        )
        
        # Case: URL with path more than / selected
        rrn2.SelectItem()
        cup_actions = _change_url_prefix_actions_in_top_level_menu(mw)
        assertEqual(
            [
                ('Set As Default Domain', True),
                ('Set As Default Directory', True),
            ],
            [
                (mi.ItemLabelText, mi.Enabled) for mi in 
                cup_actions
            ]
        )
        
        # Case: URL with path more than / selected; matches domain
        _select_change_url_prefix_action(mw, cup_actions[0])
        cup_actions = _change_url_prefix_actions_in_top_level_menu(mw)
        assertEqual(
            [
                ('Clear Default Domain', True),
                ('Set As Default Directory', True),
            ],
            [
                (mi.ItemLabelText, mi.Enabled) for mi in 
                cup_actions
            ]
        )
        
        # Case: URL with path more than / selected; matches prefix
        _select_change_url_prefix_action(mw, cup_actions[1])
        cup_actions = _change_url_prefix_actions_in_top_level_menu(mw)
        assertEqual(
            [
                ('Set As Default Domain', True),
                ('Clear Default Directory', True),
            ],
            [
                (mi.ItemLabelText, mi.Enabled) for mi in 
                cup_actions
            ]
        )


def _change_url_prefix_actions_in_top_level_menu(mw: MainWindow) -> list[wx.MenuItem]:
    entity_menu = mw.entity_menu  # cache
    entity_menu.ProcessEvent(wx.MenuEvent(type=wx.EVT_MENU_OPEN.typeId))
    entity_menu.ProcessEvent(wx.MenuEvent(type=wx.EVT_MENU_CLOSE.typeId))
    cup_actions = [
        mi for mi in entity_menu.MenuItems
        if mi.ItemLabelText in [
            'Set As Default Domain',
            'Set As Default Directory',
            'Clear Default Domain',
            'Clear Default Directory',
        ]
    ]
    return cup_actions


def _select_change_url_prefix_action(mw: MainWindow, mi: wx.MenuItem) -> None:
    entity_menu = mw.entity_menu  # cache
    entity_menu.ProcessEvent(wx.MenuEvent(type=wx.EVT_MENU.typeId, id=mi.Id, menu=None))


# ------------------------------------------------------------------------------
# Test: EntityTree: Label Tooltips

async def test_when_hover_over_resource_node_label_then_tooltip_always_contains_full_url() -> None:
    # ...even if Default URL Prefix is set
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        rr = RootResource(project, '', Resource(project, 'https://neocities.org/'))
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rrn = root_ti.find_child(rr.resource.url, project.default_url_prefix)
        
        assertIn('URL: https://neocities.org/', rrn.Tooltip('label') or '')
        await mw.entity_tree.set_default_domain_to_entity_at_tree_item(rrn)
        assertIn('URL: https://neocities.org/', rrn.Tooltip('label') or '')


async def test_when_hover_over_resource_group_node_label_then_tooltip_always_contains_full_url_pattern() -> None:
    # ...even if Default URL Prefix is set
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
        Resource(project, 'https://xkcd.com/atom.xml')
        Resource(project, 'https://xkcd.com/rss.xml')
        rg = ResourceGroup(project, 'Feed', 'https://xkcd.com/*.xml')
        
        root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
        rgn = root_ti.find_child(rg.url_pattern, project.default_url_prefix)
        
        assertIn('URL Pattern: https://xkcd.com/*.xml', rgn.Tooltip('label') or '')
        await mw.entity_tree.set_default_domain_to_entity_at_tree_item(rgn)
        assertIn('URL Pattern: https://xkcd.com/*.xml', rgn.Tooltip('label') or '')


# ------------------------------------------------------------------------------
# Test: RootNode

# (TODO: Add basic tests)


# ------------------------------------------------------------------------------
# Test: _ResourceNode

# (TODO: Add basic tests)


async def test_rn_with_error_child_retains_child_when_new_entity_added() -> None:
    with network_down():
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            
            RootResource(project, 'Domain 1', Resource(project, 'https://nosuchdomain1.com/'))
            (rrn1_ti,) = root_ti.Children
            
            rrn1_ti.Expand()
            await wait_for(first_child_of_tree_item_is_not_loading_condition(rrn1_ti))
            assert rrn1_ti.GetFirstChild() is not None
            
            RootResource(project, 'Domain 2', Resource(project, 'https://nosuchdomain2.com/'))
            (rrn1_ti, rrn2_ti) = root_ti.Children
            assert rrn1_ti.GetFirstChild() is not None  # has failed in the past
            
            rrn2_ti.Expand()
            await wait_for(first_child_of_tree_item_is_not_loading_condition(rrn2_ti))
            assert rrn1_ti.GetFirstChild() is not None
            assert rrn2_ti.GetFirstChild() is not None


# ------------------------------------------------------------------------------
# Test: RootResourceNode

@skip('not yet automated')
async def test_rrn_icon_looks_like_anchor_and_has_correct_tooltip() -> None:
    pass


@skip('not yet automated')
async def test_rrn_title_shows_rr_name_and_url() -> None:
    pass


@skip('not yet automated')
async def test_rrn_does_not_load_children_until_initially_expanded() -> None:
    pass


@skip('not yet automated')
async def test_undownloaded_rrn_has_undownloaded_badge() -> None:
    pass


@skip('not yet automated')
async def test_downloaded_fresh_rrn_has_fresh_badge() -> None:
    pass


@skip('not yet automated')
async def test_downloaded_stale_rrn_has_stale_badge() -> None:
    pass


@skip('not yet automated')
async def test_downloaded_error_rrn_has_error_badge() -> None:
    pass


async def test_given_rr_is_not_downloaded_and_project_is_read_only_when_expand_rrn_then_shows_error_node() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            # Create project
            async with (await OpenOrCreateDialog.wait_for()).create(project_dirpath) as (mw, project):
                # Create RootResource but don't download it
                r = Resource(project, home_url)
                RootResource(project, 'Home', r)
            
            # Reopen project as read-only
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath, readonly=True) as (mw, project):
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                (home_ti,) = root_ti.Children
                
                # Expand RootResourceNode and ensure it has an _ErrorNode child
                home_ti.Expand()
                await wait_for(first_child_of_tree_item_is_not_loading_condition(home_ti))
                (error_ti,) = home_ti.Children
                assert (
                    'Cannot download: Project is read only' ==
                    error_ti.Text
                )


@skip('not yet automated')
async def test_given_rr_is_not_downloaded_and_disk_is_full_when_expand_rrn_then_shows_error_node() -> None:
    pass


@skip('not yet automated')
async def test_given_rr_is_not_downloaded_and_project_has_maximum_revisions_when_expand_rrn_then_shows_error_node() -> None:
    pass


async def test_given_rr_is_downloaded_and_is_error_when_expand_rrn_then_shows_error_node() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Download revision
            with network_down():
                r = Resource(project, home_url)
                home_rr = RootResource(project, 'Home', r)
                revision_future = home_rr.download()
                while not revision_future.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
                # Wait for download to complete, including the trailing wait
                await wait_for_download_to_start_and_finish(mw.task_tree, immediate_finish_ok=True)
                
                rr = revision_future.result()
                assert DownloadErrorDict(
                    type='gaierror',
                    message='[Errno 8] nodename nor servname provided, or not known',
                ) == rr.error_dict
            
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            (home_ti,) = root_ti.Children
            
            # Expand RootResourceNode and ensure it has an _ErrorNode child
            home_ti.Expand()
            await wait_for(first_child_of_tree_item_is_not_loading_condition(home_ti))
            (error_ti,) = home_ti.Children
            assert (
                'Error downloading URL: gaierror: [Errno 8] nodename nor servname provided, or not known' ==
                error_ti.Text
            )


async def test_given_rr_is_downloaded_but_revision_body_missing_when_expand_rrn_then_shows_error_node_and_redownloads_rr() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            # Download revision
            async with (await OpenOrCreateDialog.wait_for()).create(project_dirpath) as (mw, project):
                r = Resource(project, home_url)
                home_rr = RootResource(project, 'Home', r)
                revision_future = home_rr.download()
                while not revision_future.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
                
                rr = revision_future.result()
                rr_body_filepath = rr._body_filepath  # capture
            
            # Simulate loss of revision body file, perhaps due to an
            # incomplete copy of a .crystalproj from one disk to another
            # (perhaps because of bad blocks in the revision body file)
            os.remove(rr_body_filepath)
            
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                (home_ti,) = root_ti.Children
                
                # Expand RootResourceNode and ensure it has an _ErrorNode child
                # 
                # TODO: In the future, block on the redownload finishing
                #       and list the links in the redownloaded revision,
                #       WITHOUT needing to reopen the project later
                home_ti.Expand()
                await wait_for(first_child_of_tree_item_is_not_loading_condition(home_ti))
                (error_ti,) = home_ti.Children
                assert (
                    'Cannot list links: URL revision body is missing. Recommend delete and redownload.' ==
                    error_ti.Text
                )
                
                # Wait for redownload to complete
                await wait_for_download_to_start_and_finish(mw.task_tree, immediate_finish_ok=True)
                
                # Reexpand RootResourceNode and ensure the children are the same
                home_ti.Collapse()
                home_ti.Expand()
                await wait_for(first_child_of_tree_item_is_not_loading_condition(home_ti))
                (error_ti,) = home_ti.Children
                assert (
                    'Cannot list links: URL revision body is missing. Recommend delete and redownload.' ==
                    error_ti.Text
                )
            
            # Reopen same project
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as (mw, project):
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                (home_ti,) = root_ti.Children
                
                # Expand RootResourceNode and ensure it now lists the links in
                # the redownloaded revision
                home_ti.Expand()
                await wait_for(
                    first_child_of_tree_item_is_not_loading_condition(home_ti),
                    timeout=3.0  # took 2.2s on Linux CI
                )
                children = home_ti.Children
                assert not (
                    len(children) >= 1 and 
                    'Cannot list links:' in children[0].Text
                )


# ------------------------------------------------------------------------------
# Test: NormalResourceNode

# (TODO: Add basic tests)


# ------------------------------------------------------------------------------
# Test: LinkedResourceNode

# (TODO: Add basic tests)


# ------------------------------------------------------------------------------
# Test: ClusterNode

# (TODO: Add basic tests)


# ------------------------------------------------------------------------------
# Test: ResourceGroupNode

@skip('not yet automated')
async def test_rgn_icon_looks_like_folder_and_has_correct_tooltip() -> None:
    pass


@skip('not yet automated')
async def test_rgn_title_shows_group_name_and_url() -> None:
    pass


@skip('covered by: test_given_more_node_selected_when_expand_more_node_then_first_newly_visible_child_is_selected')
async def test_rgn_does_not_load_children_until_initially_expanded() -> None:
    pass


@skip('covered by: test_given_more_node_selected_when_expand_more_node_then_first_newly_visible_child_is_selected')
async def test_rgn_only_shows_first_100_children_initially_and_has_a_more_node_showing_how_many_remain() -> None:
    pass


@skip('covered by: test_given_more_node_selected_when_expand_more_node_then_first_newly_visible_child_is_selected')
async def test_when_expand_more_node_in_rgn_then_shows_20_more_children_and_a_new_more_node() -> None:
    pass


async def test_given_more_node_selected_when_expand_more_node_then_first_newly_visible_child_is_selected() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        if True:
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Create future group members
            for i in range(1, 1000+1):
                Resource(project, comic_pattern.replace('#', str(i)))
            
            # Create group
            ResourceGroup(project, 'Comic', comic_pattern)
            
            root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
            
            comic_group_ti = root_ti.GetFirstChild()
            assert comic_group_ti is not None
            assert f'{comic_pattern} - Comic' == comic_group_ti.Text
            
            # Ensure first child of group (not displayed) is "Loading..."
            cg_child_ti = comic_group_ti.GetFirstChild()
            assert cg_child_ti is not None
            assert 'Loading...' == cg_child_ti.Text
            
            comic_group_ti.Expand()
            await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_group_ti))
            
            # Ensure after expanding that first 100 children are shown initially,
            # followed by a "# more" node
            cg_children_tis = comic_group_ti.Children
            assert len(cg_children_tis) == 100 + 1
            for i in range(0, 100):
                expected_comic_url = comic_pattern.replace('#', str(i + 1))
                assert expected_comic_url == cg_children_tis[i].Text
            more_ti = cg_children_tis[-1]
            more_ti.ScrollTo()
            assert '900 more' == more_ti.Text
            
            more_ti.Expand()
            def more_children_visible() -> bool | None:
                assert comic_group_ti is not None
                return (len(comic_group_ti.Children) > (100 + 1)) or None
            await wait_for(more_children_visible)
            
            # Ensure after expanding "# more" node that another 20 children are shown
            cg_children_tis = comic_group_ti.Children
            assert len(cg_children_tis) == 100 + 20 + 1
            for i in range(0, 100 + 20):
                expected_comic_url = comic_pattern.replace('#', str(i + 1))
                assert expected_comic_url == cg_children_tis[i].Text
            more_ti = cg_children_tis[-1]
            more_ti.ScrollTo()
            assert '880 more' == more_ti.Text
            assert False == more_ti.IsExpanded()
            
            more_ti.SelectItem()
            
            more_ti.Expand()
            def more_children_visible() -> bool | None:
                assert comic_group_ti is not None
                return (len(comic_group_ti.Children) > (100 + 20 + 1)) or None
            await wait_for(more_children_visible)
            
            # Ensure after expanding a selected "# more" node that the
            # first newly visible child inherits the selection
            cg_children_tis = comic_group_ti.Children
            assert len(cg_children_tis) == 100 + 20 + 20 + 1
            node_in_position_of_old_more_node = cg_children_tis[100 + 20]
            assert True == node_in_position_of_old_more_node.IsSelected()


async def test_given_more_node_with_large_item_count_then_displays_count_with_commas() -> None:
    # Initialize locale based on LANG='en_US.UTF-8'
    old_lang = os.environ.get('LANG')
    os.environ['LANG'] = 'en_US.UTF-8'
    old_locale = locale.setlocale(locale.LC_ALL)
    locale.setlocale(locale.LC_ALL, '')
    try:
        with served_project('testdata_xkcd.crystalproj.zip') as sp:
            # Define URLs
            if True:
                comic_pattern = sp.get_request_url('https://xkcd.com/#/')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
                # Create future group members
                for i in range(1, 1200+1):
                    Resource(project, comic_pattern.replace('#', str(i)))
                
                # Create group
                ResourceGroup(project, 'Comic', comic_pattern)
                
                root_ti = TreeItem.GetRootItem(mw.entity_tree.window)
                
                comic_group_ti = root_ti.GetFirstChild()
                assert comic_group_ti is not None
                assert f'{comic_pattern} - Comic' == comic_group_ti.Text
                
                comic_group_ti.Expand()
                await wait_for(first_child_of_tree_item_is_not_loading_condition(comic_group_ti))
                
                cg_children_tis = comic_group_ti.Children
                more_ti = cg_children_tis[-1]
                more_ti.ScrollTo()
                assert '1,100 more' == more_ti.Text
    finally:
        if old_lang is None:
            del os.environ['LANG']
        else:
            os.environ['LANG'] = old_lang
        locale.setlocale(locale.LC_ALL, old_locale)


# ------------------------------------------------------------------------------
# Test: GroupedLinkedResourcesNode

# (TODO: Add basic tests)


# ------------------------------------------------------------------------------
# Test: MorePlaceholderNode

# (TODO: Add basic tests)


# ------------------------------------------------------------------------------



import os


def tests_are_running() -> bool:
    return os.environ.get('CRYSTAL_RUNNING_TESTS', 'False') == 'True'


def set_tests_are_running() -> None:
    os.environ['CRYSTAL_RUNNING_TESTS'] = 'True'



# Tests for the following kinds of scenarios:
# 
# - Read/write of project data that has "bad blocks" or other corruption
#   detected by the underlying device driver or filesystem.
# 
# - Read of project data where the project is slightly malformed
#   in a way that is repairable. For example a ResourceRevision
#   whose body file is missing from the revisions directory.

from contextlib import redirect_stderr
from crystal.model import Project, Resource
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import served_project
from crystal.tests.util.wait import DEFAULT_WAIT_PERIOD
from crystal.tests.util.windows import OpenOrCreateDialog
import io
import os
from unittest import skip

# ------------------------------------------------------------------------------
# Test: RevisionBodyMissingError

async def test_given_default_revision_with_missing_body_when_download_related_resource_then_deletes_and_redownloads_revision() -> None:
    # Read Case: Error when reading the revision body from disk when parsing links from HTML
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            # Download revision
            r = Resource(project, home_url)
            revision_future = r.download_body()
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            
            # Simulate loss of revision body file
            revision = revision_future.result()
            os.remove(revision._body_filepath)
            
            # Download related resource
            with redirect_stderr(io.StringIO()) as captured_stderr:
                revision_future = r.download(wait_for_embedded=True, needs_result=True)
                while not revision_future.done():
                    await bg_sleep(DEFAULT_WAIT_PERIOD)
            assert 'is missing its body on disk. Redownloading it.' in captured_stderr.getvalue()
            revision = revision_future.result()
            assert revision.has_body
            with revision.open():  # ensure no error
                pass


@skip('fails: not implemented')
async def test_given_default_revision_with_missing_body_when_related_resource_served_then_serves_404_error_page_saying_revision_is_corrupt_and_requires_redownload() -> None:
    # Read Case 1: Error when reading the revision body from disk when parsing links from HTML
    # Read Case 2: Error when reading the revision body from disk when serving non-HTML revision
    pass


# ------------------------------------------------------------------------------
# Test: RevisionBodyIOError

@skip('fails: not implemented')
async def test_given_default_revision_with_body_containing_bad_blocks_when_download_related_resource_then_deletes_and_redownloads_revision() -> None:
    # Write Case: Error when downloading the revision body to disk initially
    # Read Case: Error when reading the revision body from disk when parsing links from HTML
    pass


@skip('fails: not implemented')
async def test_given_default_revision_with_body_containing_bad_blocks_when_related_html_resource_served_then_serves_404_error_page_saying_revision_is_corrupt_and_requires_redownload() -> None:
    # Read Case: Error when reading the revision body from disk when parsing links from HTML
    pass


@skip('fails: not implemented')
async def test_given_default_revision_with_body_containing_bad_blocks_when_related_non_html_resource_served_then_resets_connection_abruptly() -> None:
    # Read Case: Error when reading the revision body from disk when serving non-HTML revision
    """
    # Set SO_LINGER to 1,0 which, by convention, causes a
    # connection reset to be sent (RST) when close is called,
    # instead of the standard FIN shutdown sequence.
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack(
        'hh' if is_windows() else 'ii', 1, 0))
    sock.close()
    """


# ------------------------------------------------------------------------------


from collections.abc import Callable
from concurrent.futures import Future
from crystal.tests.util.wait import wait_for
from crystal.util.bulkheads import capture_crashes_to_stderr
from crystal.util.xthreading import bg_call_later
from typing import TypeVar

_R = TypeVar('_R')


_DEFAULT_WAIT_TIMEOUT_FOR_UNIT = 4.0

async def bg_call_and_wait(callable: Callable[[], _R], *, timeout: float | None=None) -> _R:
    """
    Start the specified callable on a background thread and
    waits for it to finish running.
    
    The foreground thread IS released while waiting, so the callable can safely
    make calls to fg_call_later() and fg_call_and_wait() without deadlocking.
    """
    if timeout is None:
        timeout = _DEFAULT_WAIT_TIMEOUT_FOR_UNIT
    
    result_cell = Future()  # type: Future[_R]
    @capture_crashes_to_stderr
    def bg_task() -> None:
        result_cell.set_running_or_notify_cancel()
        try:
            result_cell.set_result(callable())
        except BaseException as e:
            result_cell.set_exception(e)
    bg_call_later(bg_task)
    # NOTE: Releases foreground thread while waiting
    await wait_for(
        lambda: result_cell.done() or None, timeout,
        message=lambda: f'Timed out waiting for {callable}',
        stacklevel_extra=1)
    return result_cell.result()





def try_get_int(params: dict[str, list[str]], key: str) -> int | None:
    str_value = try_get_str(params, key)
    if str_value is None:
        return None
    try:
        return int(str_value)
    except ValueError:
        return None


def try_get_str(params: dict[str, list[str]], key: str) -> str | None:
    values = params.get(key)
    if values is None:
        return None
    if len(values) != 1:
        return None
    return values[0]



#!/usr/bin/env python
"""
Home of the main function, which starts the program.
"""

# TODO: Eliminate use of deprecated "annotation" future
from __future__ import annotations

# NOTE: Avoid importing anything outside the Python standard library
#       at the top-level of this module, including from the "crystal" package,
#       in case the import itself fails.
#       
#       Import failure is more common in py2app and py2exe contexts and
#       is easier to debug when it does NOT happen at the top-level.
#       
#       Therefore many imports in this file should occur directly within functions.
import argparse
import atexit
import datetime
import locale
import os
import os.path
import shutil
import sys
import threading
import time
import traceback

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from collections.abc import Callable
    from crystal.browser import MainWindow
    from crystal.model import Project
    from crystal.progress import OpenProjectProgressListener
    from crystal.shell import Shell
    from typing import Optional
    import wx


def main() -> None:
    """
    Main function. Starts the program.
    """
    _main(sys.argv[1:])


def _main(args: list[str]) -> None:
    # If running as Mac app or as Windows executable, redirect stdout and 
    # stderr to file, since these don't exist in these environments.
    # Use line buffering (buffering=1) so that prints are observable immediately.
    interactive = 'TERM' in os.environ
    log_to_file = (
        (getattr(sys, 'frozen', None) == 'macosx_app' and not interactive) or
        (getattr(sys, 'frozen', None) == 'windows_exe')
    )
    if log_to_file:
        from appdirs import user_log_dir
        from crystal import APP_AUTHOR, APP_NAME
        
        log_dirpath = user_log_dir(APP_NAME, APP_AUTHOR)
        os.makedirs(log_dirpath, exist_ok=True)
        
        sys.stdout = open(
            os.path.join(log_dirpath, 'stdout.log'), 
            'w', encoding='utf-8', buffering=1)
        sys.stderr = open(
            os.path.join(log_dirpath, 'stderr.log'), 
            'w', encoding='utf-8', buffering=1)
    
    # If CRYSTAL_FAULTHANDLER == True or running from source,
    # enable automatic dumping of Python tracebacks if wx has a segmentation fault
    if (os.environ.get('CRYSTAL_FAULTHANDLER', 'False') == 'True' or
            getattr(sys, 'frozen', None) is None):
        import faulthandler
        faulthandler.enable()
    
    # Print uncaught exceptions
    if True:
        # Print uncaught exceptions raised by Thread.run()
        def threading_excepthook(args) -> None:
            from crystal.util import cli
            
            err_file = sys.stderr
            print(cli.TERMINAL_FG_RED, end='', file=err_file)
            print('Exception in background thread:', file=err_file)
            traceback.print_exception(
                args.exc_type, args.exc_value, args.exc_traceback,
                file=err_file)
            print(cli.TERMINAL_RESET, end='', file=err_file)
            err_file.flush()
        threading.excepthook = threading_excepthook
        
        # Print uncaught exceptions raised in the main thread
        def sys_excepthook(exc_type, exc_value, exc_traceback) -> None:
            from crystal.util import cli
            
            err_file = sys.stderr
            print(cli.TERMINAL_FG_RED, end='', file=err_file)
            print('Exception in main thread:', file=err_file)
            traceback.print_exception(
                exc_type, exc_value, exc_traceback,
                file=err_file)
            print(cli.TERMINAL_RESET, end='', file=err_file)
            err_file.flush()
        sys.excepthook = sys_excepthook
        
        # Print exceptions occurring in unraisable contexts
        def sys_unraisablehook(args) -> None:
            from crystal.util import cli
            
            err_file = sys.stderr
            print(cli.TERMINAL_FG_RED, end='', file=err_file)
            print('Exception in unraisable context:', file=err_file)
            traceback.print_exception(
                args.exc_type, args.exc_value, args.exc_traceback,
                file=err_file)
            print(cli.TERMINAL_RESET, end='', file=err_file)
            err_file.flush()
        sys.unraisablehook = sys_unraisablehook
    
    # If running as Windows executable, also load .py files and adjacent
    # resources from the "lib" directory. Notably "tzinfo" data.
    if getattr(sys, 'frozen', None) == 'windows_exe':
        sys.path.insert(0, os.path.join(os.path.dirname(sys.executable), 'lib'))
    
    # If running as Windows executable, also look for command line arguments
    # in a text file in the current directory
    if getattr(sys, 'frozen', None) == 'windows_exe':
        if os.path.exists('arguments.txt'):
            with open('arguments.txt', encoding='utf-8') as f:
                args_line = f.read()
            # TODO: Consider using shlex.split() here to support quoted arguments
            args = args_line.strip().split(' ')  # reinterpret
    
    # 1. Enable terminal colors on Windows, by wrapping stdout and stderr
    # 2. Strip colorizing ANSI escape sequences when printing to a log file
    import colorama
    colorama.init()
    
    # Ensure the main package can be imported
    try:
        import crystal
    except ImportError:
        # Maybe it's in the current directory?
        sys.path.append(os.getcwd())
        try:
            import crystal
        except ImportError:
            sys.exit('Can\'t find the main "crystal" package on your Python path.')
    
    # Initialize global configuration
    from crystal.util.xurllib import patch_urlparse_to_never_raise_valueerror
    patch_urlparse_to_never_raise_valueerror()
    
    # Filter out strange "psn" argument (ex: '-psn_0_438379') that
    # macOS does sometimes pass upon first launch when run as a
    # binary downloaded from the internet.
    args = [a for a in args if not a.startswith('-psn_')]  # reinterpret
    
    from crystal.util.xos import is_linux

    # Parse CLI arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--shell',
        help='Start a CLI shell after opening a project.',
        action='store_true',
    )
    parser.add_argument(
        '--serve',
        help='Start serving the project immediately.',
        action='store_true',
    )
    parser.add_argument(
        '--cookie',
        help='HTTP Cookie header value when downloading resources.',
        type=str,
        default=None,
    )
    parser.add_argument(
        '--readonly',
        help='Whether to open the project as read-only.',
        action='store_true',
    )
    parser.add_argument(
        '--stale-before',
        help=(
            'If specified then any resource revision older than this datetime '
            'will be considered stale and a new revision will be downloaded '
            'if a download of the related resource is requested. '
            
            'Can be an ISO date like "2022-07-17", "2022-07-17T12:47:42", '
            'or "2022-07-17T12:47:42+00:00".'
        ),
        type=datetime.datetime.fromisoformat,
        default=None,
    )
    parser.add_argument(
        '--test',
        help='Run automated tests.',
        action='store',
        nargs='*',
    )
    if is_linux():
        parser.add_argument(
            '--install-to-desktop',
            help='Install this app to the desktop.',
            action='store_true',
        )
    parser.add_argument(
        'filepath',
        # NOTE: Duplicates: Project.FILE_EXTENSION, Project.OPENER_FILE_EXTENSION
        help='Optional. Path to a *.crystalproj or *.crystalopen to open.',
        type=str,
        default=None,
        nargs='?',
    )
    parsed_args = parser.parse_args(args)  # may raise SystemExit
    
    # Interpret --stale-before datetime as in local timezone if no UTC offset specified
    if parsed_args.stale_before is not None:
        from crystal.util.xdatetime import datetime_is_aware
        if not datetime_is_aware(parsed_args.stale_before):
            from tzlocal import get_localzone
            parsed_args.stale_before = parsed_args.stale_before.replace(
                tzinfo=get_localzone())  # reinterpret
    
    # Profile garbage collection
    from crystal.util.xgc import PROFILE_GC, start_profiling_gc
    if PROFILE_GC:
        start_profiling_gc()
    
    # --install-to-desktop, if requested
    if is_linux() and parsed_args.install_to_desktop:
        from crystal.install import install_to_linux_desktop_environment
        install_to_linux_desktop_environment()
        sys.exit()
    
    # Start GUI subsystem
    import wx
    import wx.richtext  # must import before wx.App object is created, according to wx.richtext module docstring
    import wx.xml  # required by wx.richtext; use explicit import as hint to py2app
    
    @atexit.register
    def on_atexit() -> None:
        # Exit process immediately, without bothering to run garbage collection
        # or other cleanup processes that can take a long time
        os._exit(getattr(os, 'EX_OK', 0))
    
    last_project = None  # type: Optional[Project]
    did_quit_during_first_launch = False
    
    # 1. Create wx.App and call app.OnInit(), opening the initial dialog
    # 2. Initialize the foreground thread
    class MyApp(wx.App):
        def __init__(self, *args, **kwargs):
            from crystal import APP_NAME
            
            self._keepalive_frame = None
            self._did_finish_launch = False
            super().__init__(*args, **kwargs)
            
            # macOS: Define app name used by the "Quit X" and "Hide X" menuitems
            self.SetAppDisplayName(APP_NAME)
        
        def OnPreInit(self):
            # (May insert debugging code here in the future)
            pass
        
        def OnInit(self):
            # If running as Mac .app, LC_CTYPE may be set to the default locale
            # instead of LANG. So copy any such locale to LANG.
            # 
            # Python documentation suggests that other variables may also
            # contain locale information: https://docs.python.org/3/library/locale.html#locale.getdefaultlocale
            if 'LANG' not in os.environ:
                for alternate_lang_var in ['LC_ALL', 'LC_CTYPE', 'LANGUAGE']:
                    if alternate_lang_var in os.environ:
                        os.environ['LANG'] = os.environ[alternate_lang_var]
                        break
            
            if sys.platform.startswith('win') and sys.version_info >= (3, 8):
                # Workaround wxPython >4.0.7 plus Python 3.8 breaking locale
                # https://discuss.wxpython.org/t/wxpython4-1-1-python3-8-locale-wxassertionerror/35168
                locale.setlocale(locale.LC_ALL, 'C')
            else:
                # Auto-detect appropriate locale based on os.environ['LANG']
                locale.setlocale(locale.LC_ALL, '')
            
            # Activate wx keepalive until self._finish_launch() is called
            self._keepalive_frame = wx.Frame(None, -1, 'Crystal')
            
            # Call self._finish_launch() after a short delay if it isn't
            # called in the meantime by MacOpenFile
            def wait_for_maybe_open_file():
                time.sleep(.2)
                
                if not self._did_finish_launch:
                    wx.CallAfter(lambda: self._finish_launch())
            thread = threading.Thread(target=wait_for_maybe_open_file, daemon=False)
            thread.start()
            
            return True
        
        def MacOpenFile(self, filepath):
            if self._did_finish_launch:
                # Ignore attempts to open additional projects if one is already open
                pass
            else:
                self._finish_launch(filepath)
        
        def _finish_launch(self, filepath: str | None=None) -> None:
            self._did_finish_launch = True
            
            try:
                nonlocal last_project
                last_project = _did_launch(parsed_args, shell, filepath)
            except SystemExit:
                nonlocal did_quit_during_first_launch
                did_quit_during_first_launch = True
                return
            except Exception:
                # wx doesn't like it when exceptions escape functions invoked
                # directly on the foreground thread. So just handle here.
                import traceback
                traceback.print_exc()
                return
            finally:
                # Deactivate wx keepalive
                self._keepalive_frame.Destroy()
    
    app = None  # type: Optional[wx.App]
    from crystal.util.xthreading import is_quitting, set_foreground_thread
    set_foreground_thread(threading.current_thread())
    try:
        # (Don't insert anything between set_foreground_thread() and MyApp())
        app = MyApp(redirect=False)
        
        # Start shell if requested
        if parsed_args.shell:
            from crystal.shell import Shell
            shell = Shell()
        else:
            shell = None
        
        # Starts tests if requested
        if parsed_args.test is not None:
            from crystal.util.xthreading import (
                bg_call_later, fg_call_later, has_foreground_thread,
            )
            assert has_foreground_thread(), (
                'Expected foreground thread to be running before starting tests, '
                'because tests expect to be able to schedule callables on '
                'the foreground thread'
            )
            
            # Immediately enter testing mode
            from crystal.util.test_mode import set_tests_are_running
            set_tests_are_running()
            
            from crystal.util.tqdm_debug import patch_tqdm_to_debug_deadlocks
            patch_tqdm_to_debug_deadlocks(on_deadlock='keep_trying')
            
            # Block until test-related modules are done loading,
            # before starting bg_task() on background thread
            from crystal.tests.index import run_tests
            from crystal.util.bulkheads import capture_crashes_to_stderr
            from crystal.util.xthreading import NoForegroundThreadError

            # NOTE: Any unhandled exception will probably call os._exit(1)
            #       before reaching this decorator.
            @capture_crashes_to_stderr
            def bg_task():
                # (Don't import anything here, because strange things can
                #  happen if the foreground thread tries to import the
                #  same new modules at the same time. Instead put
                #  any needed imports directly before starting bg_task.)
                
                is_ok = False
                try:
                    is_ok = run_tests(parsed_args.test)
                finally:
                    exit_code = 0 if is_ok else 1
                    os._exit(exit_code)
            bg_call_later(bg_task)
        
        # Run GUI
        while True:
            # Process main loop until no more windows or dialogs are open
            app.MainLoop()
            if did_quit_during_first_launch:
                break
            
            # Clean up
            if shell is not None:
                shell.detach()
            if last_project is not None:
                last_project.close()
                last_project = None
            
            # Quit?
            if is_quitting():
                break
            
            # Clear first-only launch arguments
            parsed_args.filepath = None
            
            # Re-launch, reopening the initial dialog
            last_project = _did_launch(parsed_args, shell)  # can raise SystemExit
    except SystemExit as e:
        if e.code not in [None, 0]:
            # Exit with error
            raise
        else:
            # Exit normally
            pass
    finally:
        # Stop any further events from being scheduled on the main loop
        set_foreground_thread(None)
    
    # Drain any lingering events from the main loop
    if app is not None:
        if app.HasPendingEvents():
            app.ProcessPendingEvents()
            if app.HasPendingEvents():
                print('Warning: Exiting app while some pending events still exist')


def _did_launch(
        parsed_args,
        shell: Shell | None,
        filepath: str | None=None
        ) -> Project:
    """
    Raises:
    * SystemExit -- if the user quits
    """
    from crystal.progress import CancelOpenProject, OpenProjectProgressDialog
    from crystal.util.test_mode import tests_are_running

    # If project to open was passed on the command-line, use it
    if parsed_args.filepath is not None:
        filepath = parsed_args.filepath  # reinterpret
    
    # Open/create a project
    project: Project | None = None
    window: MainWindow
    try:
        with OpenProjectProgressDialog() as progress_listener:
            # Export reference to progress_listener, if running tests
            if tests_are_running():
                from crystal import progress
                progress._active_progress_listener = progress_listener
            
            # Get a project
            project_kwargs = dict(
                readonly=parsed_args.readonly,
            )
            if filepath is None:
                # NOTE: Can raise SystemExit
                retry_on_cancel = True
                project = _prompt_for_project(progress_listener, **project_kwargs)
            else:
                # NOTE: Can raise CancelOpenProject
                retry_on_cancel = False
                project = _load_project(filepath, progress_listener, **project_kwargs)
            assert project is not None
            
            # Configure project
            project.request_cookie = parsed_args.cookie
            project.min_fetch_date = parsed_args.stale_before
            
            # Create main window
            from crystal.browser import MainWindow

            # NOTE: Can raise CancelOpenProject
            window = MainWindow(project, progress_listener)
    except CancelOpenProject:
        if project is not None:
            project.close()
        if retry_on_cancel:
            return _did_launch(parsed_args, shell, filepath)
        else:
            raise SystemExit()
    
    if shell is not None:
        shell.attach(project, window)
    
    # Start serving immediately if requested
    if parsed_args.serve:
        window.start_server()
    
    return project


def _prompt_for_project(
        progress_listener: OpenProjectProgressListener,
        **project_kwargs: object
        ) -> Project:
    """
    Raises:
    * SystemExit -- if the user quits rather than providing a project
    """
    from crystal.progress import CancelOpenProject
    from crystal.ui.BetterMessageDialog import BetterMessageDialog
    import wx
    
    def on_checkbox_clicked(event: wx.CommandEvent) -> None:
        readonly_checkbox_checked = dialog.IsCheckBoxChecked()
        create_button = dialog.FindWindow(id=wx.ID_YES)
        create_button.Enabled = not readonly_checkbox_checked
    
    dialog = BetterMessageDialog(None,
        message='Create a new project or open an existing project?',
        title='Select a Project',
        # TODO: Make it possible to check/uncheck this box by pressing the R key,
        #       and draw underline under R letter to signal that such a
        #       keyboard shortcut exists
        checkbox_label='Open as &read only',
        on_checkbox_clicked=on_checkbox_clicked,
        style=wx.YES_NO,
        yes_label='&New Project',
        no_label='&Open',
        escape_is_cancel=True,
        name='cr-open-or-create-project')
    with dialog:
        dialog.SetAcceleratorTable(wx.AcceleratorTable([
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('N'), wx.ID_YES),
            wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('O'), wx.ID_NO),
            # TODO: Cannot use a normal accelerator to toggle a checkbox.
            #       Workaround.
            #wx.AcceleratorEntry(wx.ACCEL_CTRL, ord('R'), dialog._checkbox.Id),
        ]))
        
        while True:
            from crystal.util.wx_dialog import ShowModal
            choice = ShowModal(dialog)
            
            project_kwargs = {
                **project_kwargs,
                **dict(readonly=dialog.IsCheckBoxChecked()),
            }  # reinterpret
            
            try:
                if choice == wx.ID_YES:
                    return _prompt_to_create_project(dialog, progress_listener, **project_kwargs)
                elif choice == wx.ID_NO:
                    return _prompt_to_open_project(dialog, progress_listener, **project_kwargs)
                elif choice == wx.ID_CANCEL:
                    raise SystemExit()
                else:
                    raise AssertionError()
            except CancelOpenProject:
                progress_listener.reset()
                continue


def _prompt_to_create_project(
        parent: wx.Window,
        progress_listener: OpenProjectProgressListener,
        **project_kwargs: object
        ) -> Project:
    """
    Raises:
    * CancelOpenProject -- if the user cancels the prompt early
    """
    from crystal.model import Project
    from crystal.progress import CancelOpenProject, LoadUrlsProgressDialog
    import wx
    
    dialog = wx.FileDialog(parent,
        message='',
        wildcard='*' + Project.FILE_EXTENSION,
        style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
    with dialog:
        if not dialog.ShowModal() == wx.ID_OK:
            raise CancelOpenProject()
        
        project_path = dialog.GetPath()
        if not project_path.endswith(Project.FILE_EXTENSION):
            project_path += Project.FILE_EXTENSION
    
    if os.path.exists(project_path):
        shutil.rmtree(project_path)
    return Project(project_path, progress_listener, LoadUrlsProgressDialog(), **project_kwargs)  # type: ignore[arg-type]


def _prompt_to_open_project(
        parent: wx.Window,
        progress_listener: OpenProjectProgressListener,
        **project_kwargs: object
        ) -> Project:
    """
    Raises:
    * CancelOpenProject -- if the user cancels the prompt early
    """
    from crystal.model import Project
    from crystal.progress import CancelOpenProject
    from crystal.util.wx_bind import bind
    from crystal.util.wx_dialog import position_dialog_initially, ShowModal
    from crystal.util.xos import is_linux, is_mac_os, is_windows
    import wx
    
    project_path = None  # type: Optional[str]
    
    class OpenAsDirectoryHook(wx.FileDialogCustomizeHook):
        def AddCustomControls(self, customizer: wx.FileDialogCustomize):  # override
            self.open_dir_button = customizer.AddButton('Open Directory')  # type: wx.FileDialogButton
            self.open_dir_button.Disable()
            bind(self.open_dir_button, wx.EVT_BUTTON, self._on_open_directory)
        
        def UpdateCustomControls(self) -> None:  # override
            selected_itempath = file_dialog.GetPath()
            selected_itemname = os.path.basename(selected_itempath)
            project_is_selected = (
                os.path.exists(selected_itempath) and (
                    selected_itemname.endswith(Project.FILE_EXTENSION) or
                    selected_itemname.endswith(Project.OPENER_FILE_EXTENSION)
                )
            )
            if project_is_selected:
                self.open_dir_button.Enable()
            else:
                self.open_dir_button.Disable()
        
        def _on_open_directory(self, event: wx.CommandEvent) -> None:
            nonlocal project_path
            #nonlocal file_dialog
            
            project_path = file_dialog.GetPath()
            assert not (is_mac_os() or is_windows()), (
                'wx.FileDialog.EndModal() does not dismiss dialog '
                'on macOS or Windows'
            )
            file_dialog.EndModal(wx.ID_OK)
    
    file_dialog_customize_hook = OpenAsDirectoryHook()
    file_dialog = wx.FileDialog(parent,
        message='Choose a project',
        wildcard='Projects ({wc};{wc2})|{wc};{wc2}'.format(
            # If projects appear as files, then can open directly
            wc='*' + Project.FILE_EXTENSION,
            # If projects appear as directories, then must open contained opener file
            wc2='*' + Project.OPENER_FILE_EXTENSION,
        ),
        style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
    # Offer ability to open .crystalproj directories on Linux,
    # where they were historically created without a .crystalopen file.
    # 
    # NOTE: On Windows it is possible to double-click a .crystalproj directory
    #       to open it in Crystal even if it has no .crystalopen file.
    #       So no special support for opening a .crystalproj directory
    #       needs to be provided here.
    if is_linux():
        file_dialog.SetCustomizeHook(file_dialog_customize_hook)
    with file_dialog:
        if not file_dialog.ShowModal() == wx.ID_OK:
            raise CancelOpenProject()
        if project_path is None:
            project_path = file_dialog.GetPath()
        assert project_path is not None
    del file_dialog_customize_hook  # keep hook alive until after dialog closed
    
    if not os.path.exists(project_path):
        raise AssertionError()
    if not Project.is_valid(project_path):
        _show_invalid_project_dialog()
        raise CancelOpenProject()
    
    assert '_show_modal_func' not in project_kwargs
    return _load_project(
        project_path,
        progress_listener,
        **project_kwargs)  # type: ignore[arg-type]


def _load_project(
        project_path: str,
        progress_listener: OpenProjectProgressListener,
        # NOTE: Used by automated tests
        *, _show_modal_func: Optional[Callable[[wx.Dialog], int]]=None,
        **project_kwargs: object
        ) -> Project:
    """
    Raises:
    * CancelOpenProject
    """
    from crystal.model import Project, ProjectFormatError, ProjectTooNewError
    from crystal.progress import CancelOpenProject, LoadUrlsProgressDialog
    from crystal.util.wx_dialog import position_dialog_initially
    import wx
    
    if _show_modal_func is None:
        from crystal.util.wx_dialog import ShowModal as _show_modal_func
    assert _show_modal_func is not None  # help mypy
    
    try:
        return Project(project_path, progress_listener, LoadUrlsProgressDialog(), **project_kwargs)  # type: ignore[arg-type]
    except FileNotFoundError:
        # TODO: Present this error to the user nicely
        raise
    except ProjectFormatError:
        _show_invalid_project_dialog(
            project_is_likely_corrupted=True,
            _show_modal_func=_show_modal_func)
        raise CancelOpenProject()
    except ProjectTooNewError:
        dialog = wx.MessageDialog(None,
            message=(
                'This project was created by a newer version of Crystal '
                'and cannot be opened by this version of Crystal.'
            ),
            caption='Project Too New',
            style=wx.ICON_ERROR|wx.OK,
        )
        dialog.Name = 'cr-project-too-new'
        with dialog:
            position_dialog_initially(dialog)
            _show_modal_func(dialog)
        raise CancelOpenProject()
    except CancelOpenProject:
        raise


def _show_invalid_project_dialog(
        *, project_is_likely_corrupted: bool=False,
        # NOTE: Used by automated tests
        _show_modal_func: Optional[Callable[[wx.Dialog], int]]=None,
        ) -> None:
    from crystal.util.wx_dialog import (
        position_dialog_initially, set_dialog_or_frame_icon_if_appropriate,
    )
    import wx
    
    if _show_modal_func is None:
        from crystal.util.wx_dialog import ShowModal as _show_modal_func
    assert _show_modal_func is not None  # help mypy
    
    extra = ' and may be corrupted' if project_is_likely_corrupted else ''
    
    dialog = wx.MessageDialog(None,
        message=f'The selected file or directory is not a valid project{extra}.',
        caption='Invalid Project',
        style=wx.ICON_ERROR|wx.OK,
    )
    dialog.Name = 'cr-invalid-project'
    with dialog:
        position_dialog_initially(dialog)
        _show_modal_func(dialog)


# ------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



import sys
from typing import TYPE_CHECKING

__all__ = ['NotImplemented', 'NotImplementedType']


# 1. Define NotImplementedType, the type of NotImplemented
# 2. Define NotImplemented
if sys.version_info >= (3, 10):
    from types import NotImplementedType
    NotImplemented = NotImplemented  # export
else:
    # https://github.com/python/typing/issues/684#issuecomment-548203158
    if TYPE_CHECKING:
        from enum import Enum
        class NotImplementedType(Enum):
            NotImplemented = 'NotImplemented'
        NotImplemented = NotImplementedType.NotImplemented
    else:
        NotImplementedType = type(NotImplemented)
        NotImplemented = NotImplemented  # export


import bs4
from collections.abc import Iterator
from contextlib import contextmanager
from crystal.browser import MainWindow as RealMainWindow
from crystal.doc.generic import Document, Link
from crystal.doc.html import parse_html_and_links
from crystal.doc.html.basic import BasicDocument
from crystal.doc.html.soup import FAVICON_TYPE_TITLE, HtmlDocument
from crystal.model import (
    Project, Resource, ResourceRevision, ResourceRevisionMetadata,
    RootResource,
)
from crystal.tests.util.asserts import assertEqual, assertRaises
from crystal.tests.util.controls import click_button
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import served_project
from crystal.tests.util.subtests import awith_subtests, SubtestsContext
from crystal.tests.util.wait import DEFAULT_WAIT_PERIOD
from crystal.tests.util.windows import (
    MainWindow, OpenOrCreateDialog, PreferencesDialog,
)
from io import BytesIO
import lxml.html
import os
import tempfile
from textwrap import dedent
from unittest import skip
from unittest.mock import Mock, patch
from urllib.parse import ParseResult, urljoin, urlparse

# ------------------------------------------------------------------------------
# Tests: HTML Parser Option

async def test_uses_html_parser_specified_in_preferences() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, project):
            RootResource(project, 'Home', Resource(project, home_url))
            r = Resource(project, home_url)
            
            # Ensure default HTML parser for new project is lxml
            click_button(mw.preferences_button)
            pd = await PreferencesDialog.wait_for()
            html_parser_title = pd.html_parser_field.Items[pd.html_parser_field.Selection]
            assert 'Fastest - lxml' == html_parser_title
            await pd.ok()
            
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            revision = revision_future.result()
            
            # Ensure expected HTML parser is used
            with _watch_html_parser_usage() as (lxml_parse_func, bs4_parse_func):
                (_, _, _) = revision.document_and_links()
            assert (1, 0) == (lxml_parse_func.call_count, bs4_parse_func.call_count)
            
            # Switch HTML parser
            click_button(mw.preferences_button)
            pd = await PreferencesDialog.wait_for()
            pd.html_parser_field.Selection = pd.html_parser_field.Items.index(
                'Classic - html.parser (bs4)')
            await pd.ok()
            
            # Ensure new HTML parser is used
            with _watch_html_parser_usage() as (lxml_parse_func, bs4_parse_func):
                (_, _, _) = revision.document_and_links()
            assert (0, 1) == (lxml_parse_func.call_count, bs4_parse_func.call_count)


@skip('covered by: test_uses_html_parser_specified_in_preferences')
async def test_defaults_to_lxml_html_parser_for_new_projects() -> None:
    pass


async def test_uses_html_parser_parser_for_classic_projects() -> None:
    # NOTE: The testdata project is a classic project from Crystal <=1.5.0
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        with RealMainWindow(sp.project) as rmw:
            mw = await MainWindow.wait_for()
            
            click_button(mw.preferences_button)
            pd = await PreferencesDialog.wait_for()
            try:
                html_parser_title = pd.html_parser_field.Items[pd.html_parser_field.Selection]
                assert 'Classic - html.parser (bs4)' == html_parser_title
            finally:
                await pd.ok()


@contextmanager
def _watch_html_parser_usage() -> Iterator[tuple[Mock, Mock]]:
    with patch('lxml.html.document_fromstring', wraps=lxml.html.document_fromstring) as lxml_parse_func:
        with patch('crystal.util.fastsoup.BeautifulSoup', wraps=bs4.BeautifulSoup) as bs4_parse_func:
            yield (lxml_parse_func, bs4_parse_func)


# ------------------------------------------------------------------------------
# Tests: parse_html_and_links

@skip('not yet automated')
def test_recognizes_link_to_background_image() -> None:
    pass  # type_title = 'Background Image'


@skip('not yet automated')
def test_recognizes_link_to_image_using_src() -> None:
    pass  # type_title = 'Image'


@skip('not yet automated')
def test_recognizes_link_to_iframe() -> None:
    pass  # type_title = 'IFrame'


@skip('not yet automated')
def test_recognizes_link_to_frame_when_no_frameset_present() -> None:
    pass  # type_title = 'Frame'


@skip('not yet automated')
def test_recognizes_link_to_frame_when_frameset_present() -> None:
    pass  # type_title = 'Unknown' (in doc/html/basic.py)


@skip('not yet automated')
def test_recognizes_link_to_form_image() -> None:
    pass  # type_title = 'Form Image'


@skip('not yet automated')
def test_recognizes_unknown_link_type_using_src_attribute() -> None:
    # type_title = 'Unknown Embedded (%s)' % tag_name
    # 
    # ex: <audio>, <embed>, <source>, <track>, <video>
    # via: https://www.w3schools.com/tags/att_src.asp
    pass


@skip('not yet automated')
def test_recognizes_link_to_image_using_srcset() -> None:
    pass  # type_title = 'Image'


@skip('not yet automated')
def test_recognizes_text_link() -> None:
    pass  # type_title = 'Link'


@skip('not yet automated')
def test_recognizes_link_to_stylesheet() -> None:
    # ...using rel stylesheet
    # ...using type css
    # ...using css file extension
    pass  # type_title = 'Stylesheet'


@awith_subtests
async def test_recognizes_explicit_link_to_favicon(subtests: SubtestsContext) -> None:
    with subtests.test('in general'):
        HTML_TEXT = dedent(
            '''
            <!DOCTYPE html>
            <html>
                <head>
                    <title>xkcd: Fast Radio Bursts</title>
                    <link rel="icon" href="/s/919f27.ico" type="image/x-icon"/>
                    <link rel="alternate" type="application/atom+xml" title="Atom 1.0" href="/atom.xml"/>
                    <link rel="alternate" type="application/rss+xml" title="RSS 2.0" href="/rss.xml"/>
                </head>
                <body>
                    ...
                </body>
            </html>
            '''
        ).lstrip('\n')
        assert '919f27.ico' in HTML_TEXT  # ensure has explicit favicon link
        
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            os.rmdir(project_dirpath)
            with Project(project_dirpath) as project:
                rr = ResourceRevision.create_from_response(
                    Resource(project, 'https://www.apple.com/'),
                    metadata=ResourceRevisionMetadata(
                        http_version=10,
                        status_code=200,
                        reason_phrase='OK',
                        headers=[
                            ['Content-Type', 'text/html'],
                        ],
                    ),
                    body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
                (doc, links, _) = rr.document_and_links()
                assert isinstance(doc, HtmlDocument)  # ensure using soup parser
                
                # 1. Ensure detects explicit favicon link
                # 2. Ensure does not insert implicit favicon link
                (favicon_link,) = (
                    link for link in links if
                    link.type_title == FAVICON_TYPE_TITLE and
                    link.relative_url == '/s/919f27.ico'
                )
                
                # Ensure can rewrite explicit favicon link
                favicon_link.relative_url = '/favicon2.ico'
                assert 'favicon2.ico' in str(doc)
    
    with subtests.test(using='rel shortcut icon'):  # for old IE
        HTML_TEXT = '''<link rel="shortcut icon" href="/s/919f27.ico" type="image/x-icon"/>'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = (
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/s/919f27.ico'
        )
    
    with subtests.test(using='rel icon'):
        HTML_TEXT = '''<link rel="icon" href="/s/919f27.ico" type="image/x-icon"/>'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = (
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/s/919f27.ico'
        )
    
    with subtests.test(using='rel apple-touch-icon'):
        HTML_TEXT = '''<link rel="apple-touch-icon" href="/custom_icon.png">'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = (
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/custom_icon.png'
        )
    
    with subtests.test(using='ico file extension'):
        HTML_TEXT = '''<link href="/myicon.ico" />'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = (
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/myicon.ico'
        )
    
    with subtests.test(using='png file extension'):
        HTML_TEXT = '''<link href="/myicon.png" />'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = (
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/myicon.png'
        )


# NOTE: Logic for recognizing this kind of link currently lives in
#       ResourceRevision.document_and_links() rather than inside
#       any parse_html_and_links() function.
@awith_subtests
async def test_recognizes_implicit_link_to_favicon_from_site_root(subtests: SubtestsContext) -> None:
    HTML_TEXT = dedent(
        '''
        <!DOCTYPE html>
        <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-US" lang="en-US" prefix="og: http://ogp.me/ns#" class="no-js" data-layout-name="privacy-day">
            <head>
                <meta charset="utf-8"/>
                <link rel="canonical" href="https://www.apple.com/"/>
                <link rel="alternate" href="https://www.apple.com/" hreflang="en-US"/>
                <title>Apple</title>
            </head>
            <body class="page-home ac-nav-overlap globalnav-scrim globalheader-dark nav-dark">
                ...
            </body>
        </html>
        '''
    ).lstrip('\n')
    assert 'favicon.ico' not in HTML_TEXT  # ensure no explicit favicon link
    
    FORCE_BASIC_PARSER_TEXT = '<!-- <frameset></frameset> -->'
    
    with subtests.test(parser='soup'):
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            os.rmdir(project_dirpath)
            with Project(project_dirpath) as project:
                rr = ResourceRevision.create_from_response(
                    Resource(project, 'https://www.apple.com/'),
                    metadata=ResourceRevisionMetadata(
                        http_version=10,
                        status_code=200,
                        reason_phrase='OK',
                        headers=[
                            ['Content-Type', 'text/html'],
                        ],
                    ),
                    body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
                (doc, links, _) = rr.document_and_links()
                assert isinstance(doc, HtmlDocument)  # ensure using soup parser
                
                # Ensure inserts implicit favicon link
                (favicon_link,) = (
                    link for link in links if
                    link.type_title == FAVICON_TYPE_TITLE and
                    link.relative_url == '/favicon.ico'
                )
                
                # Ensure can rewrite implicit favicon link
                # (in document from soup parser)
                favicon_link.relative_url = '/favicon2.ico'
                assert 'favicon2.ico' in str(doc)
    
    with subtests.test(parser='basic'):
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            os.rmdir(project_dirpath)
            with Project(project_dirpath) as project:
                rr = ResourceRevision.create_from_response(
                    Resource(project, 'https://www.apple.com/'),
                    metadata=ResourceRevisionMetadata(
                        http_version=10,
                        status_code=200,
                        reason_phrase='OK',
                        headers=[
                            ['Content-Type', 'text/html'],
                        ],
                    ),
                    body_stream=BytesIO((HTML_TEXT + FORCE_BASIC_PARSER_TEXT).encode('utf-8')))
                (doc, links, _) = rr.document_and_links()
                assert isinstance(doc, BasicDocument)  # ensure using basic parser
                
                # Ensure inserts implicit favicon link
                (favicon_link,) = (
                    link for link in links if
                    link.type_title == FAVICON_TYPE_TITLE and
                    link.relative_url == '/favicon.ico'
                )
                
                # Characterize known limitation:
                # Ensure CANNOT rewrite implicit favicon link
                # (in document from basic parser)
                favicon_link.relative_url = '/favicon2.ico'
                assert 'favicon2.ico' not in str(doc)


# TODO: Consider recognizing implicit favicon regardless of whether
#       at site root or not
async def test_does_not_recognize_implicit_link_to_favicon_from_outside_site_root() -> None:
    HTML_TEXT = dedent(
        '''
        <!DOCTYPE html>
        <html class="en-us amr nojs en seg-consumer us" lang="en-US">
            <head>
                <title>Apple Store Online - Apple</title>
                <meta charset="utf-8" />
                <link rel="canonical" href="https://www.apple.com/store" />
                <link rel="alternate" hreflang="en-us" href="https://www.apple.com/store" />
            </head>
            <body class="">
                ...
            </body>
        </html>
        '''
    ).lstrip('\n')
    assert 'favicon.ico' not in HTML_TEXT  # ensure no explicit favicon link
    
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        with Project(project_dirpath) as project:
            rr = ResourceRevision.create_from_response(
                Resource(project, 'https://www.apple.com/store'),
                metadata=ResourceRevisionMetadata(
                    http_version=10,
                    status_code=200,
                    reason_phrase='OK',
                    headers=[
                        ['Content-Type', 'text/html'],
                    ],
                ),
                body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
            (doc, links, _) = rr.document_and_links()
            
            # Ensure does NOT insert implicit favicon link
            () = (
                link for link in links if
                link.type_title == FAVICON_TYPE_TITLE and
                link.relative_url == '/favicon.ico'
            )


@skip('not yet automated')
def test_recognizes_link_to_preloaded_resource() -> None:
    pass  # type_title = 'Preload'


@skip('not yet automated')
def test_recognizes_unknown_link_type_using_rel_attribute() -> None:
    # type_title = 'Unknown Link (rel=%s)' % ...
    # 
    # ex: alternate, author, canonical, dns-prefetch, license,
    #     manifest, me, modulepreload, next, pingback, preconnect,
    #     prefetch, prerender, prev, privacy-policy, search,
    #     terms-of-service
    # via: https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes/rel
    pass


@skip('not yet automated')
def test_recognizes_unknown_link_type_using_href_attribute() -> None:
    # type_title = 'Unknown Href (%s)' % tag_name
    # 
    # ex: <none known>
    pass


@skip('not yet automated')
def test_recognizes_onclick_eq_location_eq_ellipsis_link() -> None:
    pass  # type_title = 'Button'


@skip('not yet automated')
def test_recognizes_reference_to_absolute_url_inside_inline_script() -> None:
    pass  # type_title = 'Script Reference'


@skip('not yet automated')
def test_recognizes_reference_to_site_relative_url_inside_inline_script() -> None:
    pass  # type_title = 'Script Reference'


@skip('not yet automated')
def test_recognizes_unknown_attribute_reference_to_absolute_url() -> None:
    pass  # type_title = 'Attribute Reference'


@skip('not yet automated')
def test_does_not_recognize_mailto_or_data_or_javascript_urls_as_links() -> None:
    pass  # see: is_unrewritable_url()


async def test_does_recognize_invalid_relative_urls_as_links() -> None:
    # Ensure test data is detected as an invalid relative URL
    # using native urllib functions
    assert hasattr(urlparse, '_super'), \
        'Expected urlparse() to be patched by patch_urlparse_to_never_raise_exception()'
    assertRaises(
        ValueError,
        lambda: urlparse._super("//*[@id='"))  # type: ignore[attr-defined]
    
    # Ensure test data is detected as an valid relative URL
    # using patched urllib functions
    assertEqual(
        ParseResult(scheme='', netloc='', path="//*[@id='", params='', query='', fragment=''),
        urlparse("//*[@id='"))
    assertEqual(
        "//*[@id='",
        urljoin('https://example.com/', "//*[@id='"))
    
    # Ensure does recognize invalid relative URL as a link in general
    HTML_TEXT = '''<a href="//*[@id='"><a href="#">'''
    (_, links) = _parse_html_and_links(HTML_TEXT)
    (bad_link, good_link) = links
    assert "//*[@id='" == bad_link.relative_url
    assert '#' == good_link.relative_url
    
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        with Project(project_dirpath) as project:
            rr = ResourceRevision.create_from_response(
                Resource(project, 'https://example.com/'),
                metadata=ResourceRevisionMetadata(
                    http_version=10,
                    status_code=200,
                    reason_phrase='OK',
                    headers=[
                        ['Content-Type', 'text/html'],
                    ],
                ),
                body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
            (doc, links, _) = rr.document_and_links()
            
            # Ensure does recognize invalid relative URL as a document link
            (bad_link, good_link, favicon_link) = links
            assert "//*[@id='" == bad_link.relative_url
            assert '#' == good_link.relative_url
            assert '/favicon.ico' == favicon_link.relative_url


async def test_recognizes_http_redirect_as_a_link() -> None:
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        with Project(project_dirpath) as project:
            rr = ResourceRevision.create_from_response(
                Resource(project, 'https://www.xkcd.com/'),
                metadata=ResourceRevisionMetadata(
                    http_version=10,
                    status_code=301,
                    reason_phrase='Moved Permanently',
                    headers=[
                        ['Location', 'https://xkcd.com/'],
                    ],
                ),
                body_stream=BytesIO(b''))
            (doc, links, _) = rr.document_and_links()
            
            # Ensure inserts redirect link
            (redirect_link,) = (
                link for link in links if
                link.type_title == 'Redirect' and
                link.relative_url == 'https://xkcd.com/' and
                link.embedded
            )


@skip('fails: not yet implemented')
def test_recognizes_meta_http_equiv_refresh_as_a_link() -> None:
    # see: https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta#examples
    pass  # type_title = 'Redirect' (proposed)


@skip('not yet automated')
def test_recognizes_links_defined_by_plugins() -> None:
    # ex: plugins_minbaker.postprocess_document_and_links()
    pass


def _parse_html_and_links(html: str) -> tuple[Document, list[Link]]:
    doc_and_links = parse_html_and_links(html.encode('utf-8'), 'utf-8', 'lxml')
    assert doc_and_links is not None
    return doc_and_links


# ------------------------------------------------------------------------------



from collections.abc import Callable
from crystal.util.bulkheads import capture_crashes_to_stderr
import wx


class Timer(wx.Timer):
    """
    Runs an action every X milliseconds.
    
    Wraps the wx.Timer API to be more usable and reliable.
    """
    
    def __init__(self, action: Callable[[], None], milliseconds: int) -> None:
        """
        Starts a timer that calls the specified action at the specified period.
        
        Raises:
        * TimerError -- if timer could not be started
        """
        super().__init__()
        self._action = action
        if not self.Start(milliseconds):
            raise TimerError('Failed to start timer')
    
    # === Operations ===
    
    def stop(self) -> None:
        """
        Stops this timer.
        
        After this timer is stopped there is no way to restart it.
        """
        self.Stop()
    
    # === Internal ===
    
    @capture_crashes_to_stderr
    def Notify(self) -> None:  # override
        self._action()


class TimerError(Exception):
    pass
